//-----------------------------------------------------------------------------
// Skip down to the next divider where IList and PList are defined.  That is 
// more for public consumption.  JList is mostly an implementation detail.  The 
// most important thing here is NodeBase.  All elements of lists are 
// descendants of NodeBase.

class JList;
class JListIterator;

// Elements of JList are descendants of NodeBase
class NodeBase
  {
private:
    friend class JList;
    friend class JListIterator;
    NodeBase *Next, *Prev;
public:
    virtual ~NodeBase() { }
  };

class JListIterator
  {
    typedef JListIterator self;

public:
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef NodeBase * link_type;
    typedef JListIterator iterator;

protected:
    friend class JList;
    link_type Cur;
public:
    JListIterator() : Cur(0) { }
    explicit JListIterator(link_type strt) : Cur(strt) { }
    JListIterator(const self &cpy) : Cur(cpy.Cur) { }

    self& operator++()    { gassert(Cur); Cur=Cur->Next; return *this; }
    self  operator++(int) { iterator tmp = *this; ++*this; return tmp; }
    self& operator--()    { gassert(Cur); Cur=Cur->Prev; return *this; }
    self  operator--(int) { iterator tmp = *this; --*this; return tmp; }

    bool operator==(const iterator& x) const { return Cur == x.Cur; }
    bool operator!=(const iterator& x) const { return Cur != x.Cur; }

    link_type get() { return Cur; }
  };

// A fast if not completely safe list
class JList
  {
    typedef JList self;

public:
    typedef NodeBase  list_node;
    typedef list_node* link_type;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef JListIterator iterator;

protected:
    list_node ptrs;

protected: // Allow descendants access to NodeBase and JListIterator
    // Also allows abuse of notation
    static NodeBase * & next(NodeBase &ln) { return ln.Next; }
    static NodeBase * & next(NodeBase *ln) { return ln->Next; }
    static NodeBase * & prev(NodeBase &ln) { return ln.Prev; }
    static NodeBase * & prev(NodeBase *ln) { return ln->Prev; }
    static NodeBase * & cur(JListIterator &li) { return li.Cur; }
    static list_node  & getptrs(JList &x) { return x.ptrs; }

public:
    JList() { next(ptrs)=prev(ptrs)=&ptrs; }
    JList(link_type strt) // strt should be a loop created with detach
      { if (strt) { next(ptrs)=strt;        prev(ptrs)=prev(strt);
                    next(prev(strt))=&ptrs; prev(strt)=&ptrs;      }
        else      { next(ptrs)=prev(ptrs)=&ptrs; } }

    void clear()
      {
        link_type pos=next(ptrs);
        while (pos!= &ptrs)
          {
            link_type nxt=next(pos);
            delete pos;
            pos=nxt;
          }
        next(ptrs)=&ptrs;
        prev(ptrs)=&ptrs;
      }
    ~JList() { clear(); }

    iterator begin() { return iterator(next(ptrs)); }
    iterator end() { return iterator(&ptrs); }
    bool empty() { return next(ptrs) == &ptrs; }

    void swap(self& x) { __STL_NAMESPACE::swap(ptrs, x.ptrs); }

// Insert before a node
    iterator insert(iterator position, link_type x)
      { return insert(cur(position), x); }
    iterator insert(link_type pos, link_type x)
      {
        next(x)=pos;
        prev(x)=prev(pos);
        prev(pos)=x;
        next(prev(x))=x;
        return iterator(x);
      }
// Insert after a node
    iterator insert_after(iterator position, link_type x)
      { return insert_after(cur(position), x); }
    iterator insert_after(link_type pos, link_type x)
      {
        next(x)=next(pos);
        prev(x)=pos;
        next(pos)=x;
        prev(next(x))=x;
        return iterator(x);
      }
    void push_front(link_type x) { insert(begin(), x); }
    void push_back(link_type x) { insert(end(), x); }

// Erase a node
    void erase(iterator position)
      { erase(cur(position)); }
    void erase(link_type pos)
      {
        gassert(!empty());
        next(prev(pos))=next(pos);
        prev(next(pos))=prev(pos);
        delete pos;
      }
// Erase an interval
    void erase(iterator first, iterator last)
      { erase(cur(first), cur(last)); }
    void erase(link_type fps, link_type lps)
      {
        if (fps!=lps)
          {
            next(prev(fps)) = next(lps);
            prev(next(lps)) = prev(fps);
            do
              {
                link_type pos=fps;
                next(fps);
                delete pos;
              } while (fps!=lps);
          }
      }
// Erase head/tail
    void pop_front() { gassert(!empty()); erase(next(ptrs)); }
    void pop_back() { gassert(!empty()); erase(prev(ptrs)); }

// Detach a node
    link_type detach(iterator position)
      { return detach(cur(position)); }
    link_type detach(link_type pos)
      {
        gassert(!empty() && pos!=&ptrs);
        next(prev(pos))=next(pos);
        prev(next(pos))=prev(pos);
        prev(pos)=pos;
        next(pos)=pos;
        return pos;
      }
// Detach an interval into a loop
    link_type detach(iterator first, iterator last)
      { return detach(cur(first), cur(last)); }
    link_type detach(link_type fps, link_type lps)
      {
        if (fps!=lps)
          {
            // Could try:
            //__STL_NAMESPACE::swap(prev(fps), prev(lps));
            //__STL_NAMESPACE::swap(next(prev(fps)), next(prev(lps)));
            // but seems slightly more sensible:
            __STL_NAMESPACE::swap(next(prev(fps)), next(prev(lps)));
            __STL_NAMESPACE::swap(prev(fps), prev(lps));
            // both should work (believe it or not)
            return fps;
          }
        else
          return 0;
      }

// Attaches a loop created with detach.  Use insert to insert an individual
// node
    void attach(iterator pos, link_type toadd)
      { attach(cur(pos), toadd); }
    void attach(link_type pos, link_type toadd)
      {
        if (toadd)
          {
            __STL_NAMESPACE::swap(next(prev(pos)), next(prev(toadd)));
            __STL_NAMESPACE::swap(prev(pos), prev(toadd));
          }
      }
// Transfers an interval (from this list or another) right before the node
// position.  Safer if from a different list because it does not handle
// position contained in or neighboring the interval.
    void transfer(iterator position, iterator first, iterator last)
      { transfer(cur(position), cur(first), cur(last)); }
    void transfer(iterator position, link_type first, link_type last)
      { transfer(cur(position), first, last); }
    void transfer(link_type position, iterator first, iterator last)
      { transfer(position, cur(first), cur(last)); }
    void transfer(link_type pos, link_type fps, link_type lps)
      {
        if (fps!=lps)
          {
            link_type bps=prev(pos);
            next(prev(fps))=lps;
            next(prev(lps))=pos;
            next(bps)=fps;
            prev(pos)=prev(lps);
            prev(lps)=prev(fps);
            prev(fps)=bps;
          }
      }
// Count the size, O(n) time!
    unsigned size()
      {
        unsigned ret=0;
        for (iterator p = begin(); p != end(); ++p)
          ++ret;
        return ret;
      }
  };

//----------------------------------------------------------------------------
// IList, an invasive doubly linked list class.  Type-safe; has efficient
// memory management.  Use the auto_ptr-like classes INodePtr & ILoop for
// holding a node or loop not in a list.  CAN'T be used with polymorphic data
// USE PList instead for that

// T should be a (public) descendant of NodeBase
template <class T> class IList;
template <class T> class PList;

//------- IListIter<T>
// point into an IList<T>

template <class T>
class IListIter : protected JListIterator
  {
    typedef IListIter self;

protected:
    typedef JListIterator parent;

public:
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef pointer link_type;
    typedef self iterator;

    friend class IList<T>;
    friend class PList<T>;

public:
    IListIter() : parent() { }
    IListIter(const self &cpy) : parent((const parent &) cpy) { }
    explicit IListIter(const parent &cpy) : parent(cpy) { }
    explicit IListIter(link_type strt) : parent((NodeBase *) strt) { }
    explicit IListIter(NodeBase *strt) : parent(strt) { }

    self& operator++()    { parent::operator++(); return *this; }
    self  operator++(int) { iterator tmp = *this; ++*this; return tmp; }
    self& operator--()    { parent::operator--(); return *this; }
    self  operator--(int) { iterator tmp = *this; --*this; return tmp; }

    bool operator==(const iterator& x) const { return Cur == x.Cur; }
    bool operator!=(const iterator& x) const { return Cur != x.Cur; }

    pointer get() { return (pointer) parent::get(); }
    pointer operator->() { gassert(get()); return get(); }
    reference operator*() { gassert(get()); return *get(); }
  };

//------- INodePtr<T>, PNodePtr<T>
// for a single node detached from a list

template <class T, class L>
class _NodePtr
  {
    typedef _NodePtr<T, L> self;

public:
    typedef T element_type;
    typedef T value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef pointer link_type;
    typedef L container_type;

private:
    mutable link_type ptr;

public:
    explicit _NodePtr(link_type strt=0) : ptr(strt) { }
    explicit _NodePtr(NodeBase *strt) : ptr((link_type) strt) { }
    _NodePtr(const self &cpy) : ptr(cpy.release()) { }
    self& operator=(const self &cpy)
      {
        if (this != &cpy)
          {
            if (ptr)
              container_type::Free(ptr);
            ptr=cpy.release();
          }
        return *this;
      }
    ~_NodePtr() { if (ptr) container_type::Free(ptr); }

    reference operator*()  const { gassert(ptr); return *ptr; }
    pointer   operator->() const { gassert(ptr); return ptr; }
    pointer   get()        const { return ptr; }
    pointer   release()    const { pointer tmp(ptr); ptr = 0; return tmp; }
  };

template <class T>
class INodePtr : public _NodePtr<T, IList<T> >
  {
    typedef INodePtr<T> self;
    typedef _NodePtr<T, IList<T> > super;
    typedef super parent;
public:
    explicit INodePtr(link_type strt=0) : parent(strt) { }
    explicit INodePtr(NodeBase *strt) : parent(strt) { }
    INodePtr(const self &cpy) : parent(cpy) { }
    self& operator=(const self &cpy)
      { return parent::operator=(cpy); }
  };

template <class T>
class PNodePtr : public _NodePtr<T, PList<T> >
  {
    typedef PNodePtr<T> self;
    typedef _NodePtr<T, PList<T> > super;
    typedef super parent;
public:
    explicit PNodePtr(link_type strt=0) : parent(strt) { }
    explicit PNodePtr(NodeBase *strt) : parent(strt) { }
    PNodePtr(const self &cpy) : parent(cpy) { }
    self& operator=(const self &cpy)
      { return parent::operator=(cpy); }
  };

//------- ILoop<T>, PLoop<T>
// contains a segment detached from a list

template <class T, class L>
class _Loop
  {
    typedef _Loop<T, L> self;

public:
    typedef T element_type;
    typedef T value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef pointer link_type;
    typedef L container_type;

private:
    mutable link_type ptr;

public:
    explicit _Loop(link_type strt=0) : ptr(strt) { }
    explicit _Loop(NodeBase *strt) : ptr((link_type) strt) { }
    _Loop(const self &cpy) : ptr(cpy.release()) { }
    self& operator=(const self &cpy)
      {
        if (this != &cpy)
          {
            if (ptr)
              container_type::FreeLoop(ptr);
            ptr=cpy.release();
          }
        return *this;
      }
    ~_Loop() { if (ptr) container_type::FreeLoop(ptr); }

    pointer   get()        const { return ptr; }
    pointer   release()    const { pointer tmp(ptr); ptr = 0; return tmp; }
  };

template <class T>
class ILoop : _Loop<T, IList<T> >
  {
    typedef ILoop<T> self;
    typedef _Loop<T, IList<T> > super;
    typedef super parent;
public:
    explicit ILoop(link_type strt=0) : parent(strt) { }
    explicit ILoop(NodeBase *strt) : parent(strt) { }
    ILoop(const self &cpy) : parent(cpy) { }
    self& operator=(const self &cpy)
      { return parent::operator=(cpy); }
  };

template <class T>
class PLoop : _Loop<T, PList<T> >
  {
    typedef PLoop<T> self;
    typedef _Loop<T, PList<T> > super;
    typedef super parent;
public:
    explicit PLoop(link_type strt=0) : parent(strt) { }
    explicit PLoop(NodeBase *strt) : parent(strt) { }
    PLoop(const self &cpy) : parent(cpy) { }
    self& operator=(const self &cpy)
      { return parent::operator=(cpy); }
  };

//------- IList<T>
// IList : invasive list, maintains freestore

// NOTE THERE ARE TWO VERSIONS
// This one has a freestore (assumes all are of type T)
// Assumes T has an operator=
// One below doesn't (for polymorphic lists)

template <class T>
class IList : protected JList
  {
protected:
    static JList freestore;

public:
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef pointer link_type;
    typedef IListIter<T> iterator;

// Memory management
    static bool GarbageCollect()
      { if (freestore.empty()) return false;
        freestore.pop_front(); return true; }
protected:
    friend class INodePtr<T>;
    static void Free(iterator position)
      { Free(position.get()); }
    static void Free(pointer pos)
      { freestore.insert(freestore.begin(), pos); }
    static void Free(iterator first, iterator last)
      { Free(first.get(), last.get()); }
    static void Free(pointer first, pointer last)
      { freestore.transfer(freestore.begin(), first, last); }
    friend class ILoop<T>;
    static void FreeLoop(pointer loop)
      { freestore.attach(freestore.begin(), loop); }
    static pointer Alloc()
      { if (freestore.empty()) return new value_type;
        else return (pointer) freestore.detach(freestore.begin()); }

public:
// Constructors/Destructors
    IList() : JList() { }
    IList(const ILoop<T> &strt) : JList(strt.release()) { }

    void clear() { Free((link_type) next(ptrs), (link_type) &ptrs); }
    ~IList() { clear(); } // calls above clear which moves nodes to freelist
                          // instead of deleting them like JList

// Iterator access
    iterator begin() { return iterator(next(ptrs)); }
    iterator end() { return iterator(&ptrs); }
    bool empty() { return JList::empty(); }
    reference front() { return *(link_type)next(ptrs); }
    reference back() { return *(link_type)prev(ptrs); }

    void swap(self& x) { __STL_NAMESPACE::swap(ptrs, getptrs(x)); }

protected:
// Insert x before a node
    iterator insert(iterator position, link_type x)
      { return iterator(JList::insert(position.get(), x)); }
    iterator insert(link_type pos, link_type x)
      { return iterator(JList::insert((NodeBase *) pos, (NodeBase *) x)); }
// Insert x after a node
    iterator insert_after(iterator pos, link_type x)
      { return iterator(JList::insert_after(pos.get(), x)); }
    iterator insert_after(link_type pos, link_type x)
      { return iterator(JList::insert_after(pos, x)); }
// Insert at head/tail
    void push_front(link_type x) { insert(begin(), x); }
    void push_back(link_type x) { insert(end(), x); }

public:
// Insert a new node before pos
    iterator add(iterator pos)
      { return iterator(JList::insert(pos.get(), Alloc())); }
    iterator add(link_type pos)
      { return iterator(JList::insert(pos, Alloc())); }
// Insert a new node after pos
    iterator add_after(iterator pos)
      { return iterator(JList::insert_after(pos.get(), Alloc())); }
    iterator add_after(link_type pos)
      { return iterator(JList::insert_after(pos, Alloc())); }
// Insert a new node at the head/tail
    iterator push_front() { return add((link_type) next(ptrs)); }
    iterator push_back() { return add((link_type) &ptrs); }

// Erase a node
    void erase(iterator pos)
      { gassert(!empty()); Free((link_type) JList::detach(pos.get())); }
    void erase(link_type pos)
      { gassert(!empty()); Free((link_type) JList::detach(pos)); }
// Erase an interval
    void erase(iterator first, iterator last)
      { Free(first, last); }
    void erase(link_type first, link_type last)
      { Free(first, last); }
// Erase head/tail
    void pop_front() { erase((link_type) next(ptrs)); }
    void pop_back() { erase((link_type) prev(ptrs)); }

// Detach a node
    INodePtr<T> detach(iterator pos)
      { gassert(!empty()); return INodePtr<T>(JList::detach(pos.get())); }
    INodePtr<T> detach(link_type pos)
      { gassert(!empty()); return INodePtr<T>(JList::detach(pos)); }
// Detach an interval into a loop
    ILoop<T> detach(iterator first, iterator last)
      { return ILoop<T>(JList::detach(first.get(), last.get())); }
    ILoop<T> detach(link_type first, link_type last)
      { return ILoop<T>(JList::detach(first, last)); }

// Attaches a loop created with detach.
    void attach(iterator position, ILoop<T> toadd)
      { JList::attach(position.get(), toadd.release()); }
    void attach(link_type position, ILoop<T> toadd)
      { JList::attach(position, toadd.release()); }
// Attaches a node created with detach.
    void attach(iterator position, INodePtr<T> toadd)
      { insert(position.get(), toadd.release()); }
    void attach(link_type position, INodePtr<T> toadd)
      { insert(position, toadd.release()); }

// Transfer an interval from another list
    void transfer(iterator position, iterator first, iterator last)
      { JList::transfer(position.get(), first.get(), last.get()); }
    void transfer(iterator position, link_type first, link_type last)
      { JList::transfer(position.get(), first, last); }
    void transfer(link_type position, iterator first, iterator last)
      { JList::transfer(position, first.get(), last.get()); }
    void transfer(link_type position, link_type first, link_type last)
      { JList::transfer(position, first, last); }

// Count the size, O(n) time!
    unsigned size() { return JList::size(); }
  };


//------- PList<T>
// PList : Invasive list, no freestore, descendants of T allowed
// does NOT assume an operator =

template <class T>
class PList : protected JList
  {
public:
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef pointer link_type;
    typedef IListIter<T> iterator;
    typedef PList<T> self;
    typedef PNodePtr<T> nodeptr;
    typedef PLoop<T> loop;
    friend class PNodePtr<T>;
    friend class PLoop<T>;

// Memory management
protected:
    static void Free(iterator position)
      { Free(position.get()); }
    static void Free(pointer pos)
      { delete pos; }
    static void Free(iterator first, iterator last)
      { Free(first.get(), last.get()); }
    static void Free(pointer first, pointer last)
      { while (first!=last)
          { pointer n=static_cast<pointer>(next(first));
            delete first;
            first=n;
      }   }
    static void FreeLoop(pointer loop)
      {
        pointer cur=loop;
        do
          {
            pointer save=cur;
            cur=static_cast<pointer>(next(cur));
            Free(save);            
          } while (cur!=loop);
      }
//    static pointer Alloc()
//      { return new value_type; }

public:
// Constructors/Destructors
    PList() : JList() { }
    PList(const loop &strt) : JList(strt.release()) { }

    ~PList() { } 

// Iterator access
    iterator begin() { return iterator(next(ptrs)); }
    iterator end() { return iterator(&ptrs); }
    bool empty() { return JList::empty(); }
    reference front() { return *(link_type)next(ptrs); }
    reference back() { return *(link_type)prev(ptrs); }

    void swap(self& x) { __STL_NAMESPACE::swap(ptrs, getptrs(x)); }

protected:
// Insert x before a node
    iterator insert(iterator position, link_type x)
      { return iterator(JList::insert(position.get(), x)); }
    iterator insert(link_type pos, link_type x)
      { return iterator(JList::insert((NodeBase *) pos, (NodeBase *) x)); }
// Insert x after a node
    iterator insert_after(iterator pos, link_type x)
      { return iterator(JList::insert_after(pos.get(), x)); }
    iterator insert_after(link_type pos, link_type x)
      { return iterator(JList::insert_after(pos, x)); }
// Insert at head/tail
//    void push_front(link_type x) { insert(begin(), x); }
//    void push_back(link_type x) { insert(end(), x); }

public:
// Insert a new node before pos
    iterator add(iterator pos, pointer toadd)
      { return iterator(JList::insert(pos.get(), toadd)); }
    iterator add(link_type pos, pointer toadd)
      { return iterator(JList::insert(pos, toadd)); }
// Insert a new node after pos
    iterator add_after(iterator pos, pointer toadd)
      { return iterator(JList::insert_after(pos.get(), toadd)); }
    iterator add_after(link_type pos, pointer toadd)
      { return iterator(JList::insert_after(pos, toadd)); }
// Insert a new node at the head/tail
    iterator push_front(pointer toadd) { return add((link_type) next(ptrs), toadd); }
    iterator push_back(pointer toadd) { return add((link_type) &ptrs, toadd); }

// Erase a node
    void erase(iterator pos)
      { gassert(!empty()); Free((link_type) JList::detach(pos.get())); }
    void erase(link_type pos)
      { gassert(!empty()); Free((link_type) JList::detach(pos)); }
// Erase an interval
    void erase(iterator first, iterator last)
      { Free(first, last); }
    void erase(link_type first, link_type last)
      { Free(first, last); }
// Erase head/tail
    void pop_front() { erase((link_type) next(ptrs)); }
    void pop_back() { erase((link_type) prev(ptrs)); }

// Detach a node
    nodeptr detach(iterator pos)
      { gassert(!empty()); return nodeptr(JList::detach(pos.get())); }
    nodeptr detach(link_type pos)
      { gassert(!empty()); return nodeptr(JList::detach(pos)); }
// Detach an interval into a loop
    loop detach(iterator first, iterator last)
      { return loop(JList::detach(first.get(), last.get())); }
    loop detach(link_type first, link_type last)
      { return loop(JList::detach(first, last)); }

// Attaches a loop created with detach.
    void attach(iterator position, loop toadd)
      { JList::attach(position.get(), toadd.release()); }
    void attach(link_type position, loop toadd)
      { JList::attach(position, toadd.release()); }
// Attaches a node created with detach.
    void attach(iterator position, nodeptr toadd)
      { insert(position.get(), toadd.release()); }
    void attach(link_type position, nodeptr toadd)
      { insert(position, toadd.release()); }

// Transfer an interval from another list
    void transfer(iterator position, iterator first, iterator last)
      { JList::transfer(position.get(), first.get(), last.get()); }
    void transfer(iterator position, link_type first, link_type last)
      { JList::transfer(position.get(), first, last); }
    void transfer(link_type position, iterator first, iterator last)
      { JList::transfer(position, first.get(), last.get()); }
    void transfer(link_type position, link_type first, link_type last)
      { JList::transfer(position, first, last); }

// Count the size, O(n) time!
    unsigned size() { return JList::size(); }
  };

