#include "profiler/profiler_lowlevel.h"
#include "profiler/profiler_highlevel.h"

// Definitions for the program zones we will use in
// this file (there are also some zones in the
// 'profiler' subdirectory and in entity.cpp)

Define_Zone(draw_scene);
Define_Zone(ai_pathfind);
Define_Zone(stream_music);


void draw_scene() {
    Profile_Scope(draw_scene);

    double now = app_shell->get_time();
    double dt;
    static double last_time = 0;
    if (last_time == 0) {
        last_time = now;
    }

    dt = now - last_time;
    last_time = now;

    /* SNIP */

    profile_tracker->update();
    profile_tracker->draw(0, 300);
}


void app_init() {

    profile_tracker = new Profile_Tracker();
}


static LARGE_INTEGER get_time_reading() {
    LARGE_INTEGER freq;
    LARGE_INTEGER time;

    BOOL ok = QueryPerformanceFrequency(&freq);
    assert(ok == TRUE);

    freq.QuadPart = freq.QuadPart / 1000;

    ok = QueryPerformanceCounter(&time);
    assert(ok == TRUE);

    time.QuadPart = time.QuadPart / freq.QuadPart;

	return time;
}


bool App_Shell::init(int width, int height) {
	// Init time.
	global_base_time = get_time_reading();
	return true;
}

double App_Shell::get_time() {
    LARGE_INTEGER time = get_time_reading();
	time.QuadPart = time.QuadPart - global_base_time.QuadPart;

	return (double)time.QuadPart * 0.001;
}

int AppPaint(HWND, HDC) {
    Profile_Scope(app_shell_main_loop);

    float time_0 = app_shell->get_time();

    /* snip */

    float time_1 = app_shell->get_time();
    float dt = time_1 - time_0;


    return TRUE;
}


void Profile_Tracker::update() {
    Profile_Scope(profile_tracker_update);

    // Precompute the time factors

    double now = app_shell->get_time();

    if (update_index == 0) {
        dt = FRAME_TIME_INITIAL;
    } else {
        dt = now - last_update_time;
    }

//    if (dt < FRAME_TIME_INITIAL) dt = FRAME_TIME_INITIAL;

    last_update_time = now;
    

    int i;
    for (i = 0; i < NUM_PROFILE_TRACKER_HISTORY_SLOTS; i++) {
        precomputed_factors[i] = pow(0.1f, dt / times_to_reach_90_percent[i]);
    }

    precomputed_factors[0] = 0; // For now we want this to be instantaneous.

    Profiling_Int64 timestamp_delta;
    os_fast_get_integer_timestamp(&current_integer_timestamp);
    if (update_index == 0) {
        double sum = 0;
        for (i = 0; i < num_active_zones; i++) {
            sum += Profiling::zone_pointers_by_index[i]->total_self_ticks;
        }

        if (sum == 0) sum = 1;

        timestamp_delta = sum;
    } else {
        timestamp_delta = current_integer_timestamp - last_integer_timestamp;
        if (timestamp_delta == 0) timestamp_delta = 1;
    }

    last_integer_timestamp = current_integer_timestamp;

    double timestamps_per_second = (double)timestamp_delta / dt;
    if (update_index < NUM_THROWAWAY_UPDATES) {
        integer_timestamps_per_second.eternity_set(timestamps_per_second);
    } else {
        integer_timestamps_per_second.update(timestamps_per_second, precomputed_factors);
    }

    double timestamps_to_seconds;
    if (timestamps_per_second) {
        timestamps_to_seconds = 1.0 / timestamps_per_second;
    } else {
        timestamps_to_seconds = 0;
    }

    for (i = 0; i < MAX_PROFILING_ZONES; i++) {
        Program_Zone *zone = Profiling::zone_pointers_by_index[i];
        if (zone == NULL) break;

        Profile_Tracker_Data_Record *record = &data_records[i];

        record->index = zone->index;

        double self_time = zone->total_self_ticks * timestamps_to_seconds;
        double hier_time = zone->total_hier_ticks * timestamps_to_seconds;
        
        double entry_count = zone->total_entry_count;

        assert(hier_time >= 0);

        if (update_index < NUM_THROWAWAY_UPDATES) {
            int j;
            for (j = 0; j < NUM_PROFILE_TRACKER_HISTORY_SLOTS; j++) {
                record->self_time.values[j] = self_time;
                record->self_time.variances[j] = self_time;
                record->hierarchical_time.values[j] = hier_time;
                record->hierarchical_time.variances[j] = hier_time;
                record->entry_count.values[j] = entry_count;
                record->entry_count.variances[j] = entry_count;
            }
        } else {
            // @Improvement:  If we want to be very special, we can verify here that
            // our self_time and hier_times roughly add up to dt.  If they don't, then
            // maybe SpeedStep kicked in, or something else happened to disrupt the
            // accuracy of our integer timestamps, and they should not be trusted
            // this frame...
            record->self_time.update(self_time, precomputed_factors);
            record->hierarchical_time.update(hier_time, precomputed_factors);
            record->entry_count.update(entry_count, precomputed_factors);
        }

        // XXX This should probably be cleared out only by a core system thing
        // that runs once per frame (i.e. we should be able to instantiate
        // multiple profilers looking at different things and they should all
        // work)
        zone->total_self_ticks = 0;
        zone->total_hier_ticks = 0;
        zone->total_entry_count = 0;
    }

    frame_time.update(dt, precomputed_factors);

    update_index++;


    int slot = IMPORTANT_SLOT;

    int cursor = 0;
    for (i = 0; i < MAX_PROFILING_ZONES; i++) {
        Profile_Tracker_Data_Record *record = &data_records[i];
        if (record->index == -1) continue;

        switch (displayed_quantity) {
        case SELF_TIME:
            record->displayed_quantity = record->self_time.values[slot] * 1000;
            break;
        case SELF_STDEV: {
            float stdev = get_stdev(&record->self_time, slot);
            record->displayed_quantity = stdev * 1000;
            break;
        }
        case HIERARCHICAL_TIME:
            record->displayed_quantity = record->hierarchical_time.values[slot] * 1000;
            break;
        case HIERARCHICAL_STDEV: {
            float stdev = get_stdev(&record->hierarchical_time, slot);
            record->displayed_quantity = stdev * 1000;
            break;
        }
        default:
            assert(0);
        }

        sorted_pointers[cursor++] = record;
    }

    num_active_zones = cursor;

    qsort(sorted_pointers, num_active_zones, sizeof(Profile_Tracker_Data_Record *), sort_records);

    float total_self_time[NUM_PROFILE_TRACKER_HISTORY_SLOTS];

    int j;
    for (j = 0; j < NUM_PROFILE_TRACKER_HISTORY_SLOTS; j++) {
        float sum = 0;
        for (i = 0; i < num_active_zones; i++) {
            sum += sorted_pointers[i]->self_time.values[j];
        }

        if (sum == 0) sum = 0.01;
        total_self_time[j] = sum;
    }
}
