#include "contourmd.h"
#include "contour.h"
#include "int2dmd.h"
#include "int2d.h"
#include "int2dln.h"
#include "int2dpr.h"
#include "meshbtt.h"
#include "root2d.h"
#include <stdio.h>
#include "makeps.h"
#include "jfstr.h"
#include "draw.h"
#include "blow_profile/profiler.h"

#if defined(PROFILE_HOOKS) && defined(PROFILE_2D)
Define_Zone(contour_once);
  Define_Zone(mesh_refinement);
  Define_Zone(process_results);
    Define_Zone(interpolate);
    Define_Zone(contour_interpolant);
    Define_Zone(puzzle_piece);
    Define_Zone(puzzle_finish);
#endif

static ScalarType epsilon = 0.00000005;

typedef Interpolant2DMaxD InterpolantTypeMaxD;
typedef Interpolant2D InterpolantTypeReg;
typedef Interpolant2DPrecise InterpolantTypePr;
typedef InterpolantLinear2D InterpolantTypeLinear;
typedef InterpolantTypeMaxD::TraitsType TraitsMaxD;
typedef InterpolantTypeReg::TraitsType TraitsReg;
typedef InterpolantTypePr::TraitsType TraitsPr;
typedef InterpolantTypeLinear::TraitsType TraitsLinear;
typedef ContourMaxD<BTTMesh<TraitsMaxD>, InterpolantTypeMaxD> ContourTypeMaxD;
typedef Contour<BTTMesh<TraitsReg>, InterpolantTypeReg> ContourTypeReg;
typedef Contour<BTTMesh<TraitsPr>, InterpolantTypePr> ContourTypePr;
typedef Contour<BTTMesh<TraitsLinear>, InterpolantTypeLinear> ContourTypeLinear;
typedef Root2D::TriangularCubicUnit CubicUnit;
typedef std::list<Root2D::Contour> ContourList;
typedef ContourList::iterator ContourIter;

// defined in root2d_test.cpp
void LinearCubic(CubicUnit &cubic, ScalarType nx, ScalarType ny, ScalarType offset);

struct ErrorTuple
  {
    ScalarType mError;
    ScalarType mValue;
    ScalarType mGrad;
    ScalarType mX;
    ScalarType mY;
  };

struct ErrorData
  {
    ErrorTuple mMaxError, mMaxValue;
    ScalarType mAvgError;
    ScalarType mAvgValue;
  };

struct FunctionProfiled : public Function2DMaxD
  {
    int mFCount;
    int mGCount;

    void ResetCounts()
      {
        mFCount=0;
        mGCount=0;
      }
  };

struct CubicToContour : public FunctionProfiled 
  {
    Root2D::TriangularCubicUnit  mCubic;
    Root2D::TriangularQuadraticUnit  mGradients[2];
    void Init()
      {
        mCubic.PartialDerivativeS(mGradients[0]);
        mCubic.PartialDerivativeT(mGradients[1]);
      }

    virtual ScalarType Value(ScalarType x, ScalarType y)
      {
        ScalarType st[2];
        st[0]=x;
        st[1]=y;
        ++mFCount;
        return mCubic.EvaluateAt(st);
      }
    void Grad(ScalarType x, ScalarType y, 
              ScalarType &dx, ScalarType &dy)
      {
        ScalarType st[2];
        st[0]=x;
        st[1]=y;
        ++mGCount;
        dx=mGradients[0].EvaluateAt(st);
        dy=mGradients[1].EvaluateAt(st);
      }
    virtual void ValueAndGrad(ScalarType x, ScalarType y, ScalarType &value,
                              ScalarType &gradx, ScalarType &grady)
      {
        ScalarType st[2];
        st[0]=x;
        st[1]=y;
        ++mFCount;
        ++mGCount;
        value = mCubic.EvaluateAt(st);
        gradx = mGradients[0].EvaluateAt(st);
        grady = mGradients[1].EvaluateAt(st);
      }
    virtual ScalarType MaxDeriv(ScalarType x, ScalarType y, ScalarType r)
      {
        Root2D::TriangularQuadraticUnit  subgradients[4];
        ScalarType st[2][2];
        st[0][0]=x-r;
        st[0][1]=y-r;
        st[1][0]=x+r;
        st[1][1]=y+r;
        ScalarType ret=0;
        for (int i=0; i<2; ++i)
          {
            mGradients[i].Subdivide(st, &subgradients[2*i]);
            ScalarType maxv=0;
            for (int j=2*i; j<2*i+2; ++j)
              for (int k=0; k<6; ++k)
                {
                  ScalarType curv=subgradients[j].mY[k];
                  curv*=curv;
                  maxv=Matrix::Max(curv, maxv);
                }
              ret+=maxv;
          }

        return sqrt(ret);
      }
  };

class BiCubicToContour : public FunctionProfiled 
  {
protected:
    static ScalarType EvaluateCubicBezier(ScalarType coeff[4], ScalarType t)
      {
        Root1D::BezierUnit *evilcast=
            reinterpret_cast<Root1D::BezierUnit *>(&coeff[0]);
        return evilcast->EvaluateAt(t);
      }

    static ScalarType EvaluateBiCubic(ScalarType coeff[4][4], 
                                      ScalarType x, ScalarType y)
      {
        ScalarType temp[4];
        for (int i=0; i<4; ++i)
          temp[i]=EvaluateCubicBezier(coeff[i], x);
        return EvaluateCubicBezier(temp, y);
      }
    ScalarType mGrad[2][4][4];
    ScalarType mMaxDeriv;

public:
    ScalarType mCoeff[4][4];
    void Init()
      {
        // find successive differences of mCoeff
        // degree raise back up to cubic
        // store in mGrad
        // also compute mMaxDeriv (though we could
        // be more accurate if we implement a subdivide procedure)
        int i, j;
        ScalarType minmaxd[2][2];

        // mGrad[0] is partial derivative in x
        for (i=0; i<4; ++i)
          {
            ScalarType temp[3];
            for (j=0; j<3; ++j)
              temp[j]=3*(mCoeff[i][j+1]-mCoeff[i][j]);
            mGrad[0][i][0]=temp[0];
            mGrad[0][i][1]=temp[0]+(temp[1]-temp[0])*(2./3);
            mGrad[0][i][2]=temp[2]+(temp[1]-temp[2])*(2./3);
            mGrad[0][i][3]=temp[2];
            if (i==0)
              minmaxd[0][0]=minmaxd[1][0]=temp[0];
            for (j=0; j<4; ++j)
              if (mGrad[0][i][j]<minmaxd[0][0])
                minmaxd[0][0]=mGrad[0][i][j];
              else if (mGrad[0][i][j]>minmaxd[1][0])
                minmaxd[1][0]=mGrad[0][i][j];
          }
        // mGrad[1] is partial derivative in y
        for (i=0; i<4; ++i)
          {
            ScalarType temp[3];
            for (j=0; j<3; ++j)
              temp[j]=3*(mCoeff[j+1][i]-mCoeff[j][i]);
            mGrad[1][0][i]=temp[0];
            mGrad[1][1][i]=temp[0]+(temp[1]-temp[0])*(2./3);
            mGrad[1][2][i]=temp[2]+(temp[1]-temp[2])*(2./3);
            mGrad[1][3][i]=temp[2];
            if (i==0)
              minmaxd[0][1]=minmaxd[1][1]=temp[0];
            for (j=0; j<4; ++j)
              if (mGrad[1][j][i]<minmaxd[0][1])
                minmaxd[0][1]=mGrad[1][j][i];
              else if (mGrad[1][j][i]>minmaxd[1][1])
                minmaxd[1][1]=mGrad[1][j][i];
          }
        for (i=0; i<2; ++i)
          for (int j=0; j<2; ++j)
            minmaxd[i][j]=Matrix::Abs(minmaxd[i][j]);
        for (i=0; i<2; ++i)
          {
            minmaxd[0][i]=Matrix::Max(minmaxd[0][i], minmaxd[1][i]);
            minmaxd[0][i]*=minmaxd[0][i];
          }
        mMaxDeriv = sqrt(minmaxd[0][0]+minmaxd[0][1]);
      }

    virtual ScalarType Value(ScalarType x, ScalarType y)
      {
        ++mFCount;
        return EvaluateBiCubic(mCoeff, x, y);
      }
    void Grad(ScalarType x, ScalarType y, 
              ScalarType &dx, ScalarType &dy)
      {
        ++mGCount;
        dx=EvaluateBiCubic(mGrad[0], x, y);
        dy=EvaluateBiCubic(mGrad[1], x, y);
      }
    virtual void ValueAndGrad(ScalarType x, ScalarType y, ScalarType &value,
                              ScalarType &gradx, ScalarType &grady)
      {
        ++mFCount;
        ++mGCount;
        value = EvaluateBiCubic(mCoeff, x, y);
        gradx = EvaluateBiCubic(mGrad[0], x, y);
        grady = EvaluateBiCubic(mGrad[1], x, y);
      }
    virtual ScalarType MaxDeriv(ScalarType x, ScalarType y, ScalarType r)
      {
        return mMaxDeriv;
      }
  };


/*  Franke function
f := (x,y) -> 3./4.*exp( -(SQ(9*x-2) + SQ(9*y-2))/4. )
+ 3./4.*exp( -SQ(9*x+1)/49. - SQ(9*y+1)/10.)
+ 1./2.*exp(-( SQ(9*x-7) + SQ(9*y-3) )/4. )
- 1./5.*exp(-SQ(9*x-4) - SQ(9*y-7));
SQ := x->x^2;
lprint(diff(f(x,y),x));
.75*(-40.5*x+9)*exp(-.25*(9*x-2)^2-.25*(9*y-2)^2)
+.75*(-3.306122450*x-.3673469389)*exp(-.02040816327*(9*x+1)^2-.1*(9*y+1)^2)
+.5*(-40.5*x+31.5)*exp(-.25*(9*x-7)^2-.25*(9*y-3)^2)
-.2*(-162*x+72)*exp(-(9*x-4)^2-(9*y-7)^2)
lprint(diff(f(x,y),y));
.75*(-40.5*y+9.)*exp(-.250*(9*x-2)^2-.25*(9*y-2)^2)
+.75*(-16.2*y-1.8)*exp(-.02040816327*(9*x+1)^2-.1*(9*y+1)^2)
+.5*(-40.5*y+13.5)*exp(-.25*(9*x-7)^2-.25*(9*y-3)^2)
-.2*(-162*y+126)*exp(-(9*x-4)^2-(9*y-7)^2)
*/
void PrintContours(ContourList &results)
  {
    for (ContourIter cur=results.begin(); cur!=results.end(); ++cur)
      {
        if ((*cur).mEndingSides[0]==-1)
          {
            assert((*cur).mEndingSides[1]==-1);
            printf("Closed loop\n");
          }
        else
          printf("Topologically a line\n");

        for (int i=0; i<(*cur).mControl.size(); ++i)
          {
            if (i)
              if ((i%3)==0)
                printf("\n - ");
              else
                printf(" - ");
            printf("(%f,%f)", (*cur).mControl[i](0,0), 
                              (*cur).mControl[i](0,1));
          }
        printf("\n");
      }
  }

static void EpsContours(FunctionLinear2D *cubic, 
                        ContourList &results, 
                        const char *filename, bool is_cubic)
  {
    using namespace PS;
    Drawing drawing;
    State state(drawing.GetState());
    ContourIter cur;

    // draw signs of function
    //State minus(state);
    //State plus(state);
    //minus.SetRGB(1,0.7,0.8);
    //plus.SetRGB(0.7,0.8,1);
    //double step=0.05;
    //for (double x=0.0; x<=.5; x += step)
    //  for (double y=0.0; y<=.5; y += step)
    //    {
    //      Path path;
    //      path.MoveTo(ScaleUp(x-step/2), ScaleUp(y-step/2));
    //      path.LineTo(ScaleUp(x+step/2), ScaleUp(y-step/2));
    //      path.LineTo(ScaleUp(x+step/2), ScaleUp(y+step/2));
    //      path.LineTo(ScaleUp(x-step/2), ScaleUp(y+step/2));
    //      path.Close();
    //      if (cubic->Value(x,y) > 0)
    //        drawing.FillPath(path, plus);
    //      else
    //        drawing.FillPath(path, minus);
    //    }

    // draw domain
    //state.SetGray(0.7);
    //{
    //  Path path;
    //  path.MoveTo(ScaleUp(0), ScaleUp(0));
    //  path.LineTo(ScaleUp(1), ScaleUp(0));
    //  path.LineTo(ScaleUp(1), ScaleUp(1));
    //  path.LineTo(ScaleUp(0), ScaleUp(1));
    //  path.Close();
    //  drawing.StrokePath(path, state);
    //}

#if 0
    if (is_cubic)
      {
/*
        // draw joins
        state.SetGray(0.4);
        //state.SetRGB(1,0,1);
        for (cur=results.begin(); cur!=results.end(); ++cur)
          {
            Path path;
            if ((*cur).mEndingSides[0]!=-1)
              {
                path.Circle(ScaleUp((*cur).mControl[0](0,0)), 
                            ScaleUp((*cur).mControl[0](0,1)), 2);
                int last=(*cur).mControl.size()-1;
                path.Circle(ScaleUp((*cur).mControl[last](0,0)), 
                            ScaleUp((*cur).mControl[last](0,1)), 2);
              }
            drawing.FillPath(path, state);
          }
*/
        // draw contoured curve itself
        //state.SetRGB(1,0,0);
        state.SetGray(0);
        state.SetLineThickness(0.5);
        DrawSplineContour(results, drawing, state);
      }
    else
      {
        state.SetGray(0);
        state.SetLineThickness(0.5);
        DrawLinearContour(results, drawing, state);
      }
#else
    // draw control polygon
    if (is_cubic)
      {
        state.SetGray(0.6);
        state.SetLineThickness(0.15);
      }
    else
      {
        state.SetGray(0);
        state.SetLineThickness(0.5);
      }
    DrawLinearContour(results, drawing, state);

    if (is_cubic)
      {
        // draw vertices
        state.SetGray(0);
        DrawVertices(results, 0.5, drawing, state);

/*
        // draw joins
        state.SetGray(0.4);
        //state.SetRGB(1,0,1);
        for (cur=results.begin(); cur!=results.end(); ++cur)
          {
            Path path;
            if ((*cur).mEndingSides[0]!=-1)
              {
                path.Circle(ScaleUp((*cur).mControl[0](0,0)), 
                            ScaleUp((*cur).mControl[0](0,1)), 2);
                int last=(*cur).mControl.size()-1;
                path.Circle(ScaleUp((*cur).mControl[last](0,0)), 
                            ScaleUp((*cur).mControl[last](0,1)), 2);
              }
            drawing.FillPath(path, state);
          }
*/
        // draw contoured curve itself
        //state.SetRGB(1,0,0);
        state.SetGray(0);
        state.SetLineThickness(0.5);
        DrawSplineContour(results, drawing, state);
      }
#endif

    // draw ideal
    //state.SetRGB(0,1,0);
    //  {
    //    Path path;
    //    path.Circle(ScaleUp(.25), ScaleUp(.25), Radius(.2));
    //    drawing.StrokePath(path, state);
    //  }

    string epsdata;
    drawing.WriteEPS(epsdata, "Josh Levenberg", "2D Contour");

    ofstream outto(filename);
    outto << epsdata;
    outto.close();
  }

void ComputeErrorSilent(FunctionCubic2D *cubic, ContourList &results, 
                        bool is_cubic, ErrorData *ret)
  {
    ContourIter cur;
    ret->mMaxError.mError=-1;
    ret->mMaxValue.mValue=-1;
    ScalarType totalValue=0, totalError=0, totalWeight=0;

    int inc = is_cubic ? 3 : 1;

    for (cur=results.begin(); cur!=results.end(); ++cur)
      {
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> DomainType;
        DomainType lastpos=(*cur).mControl[0];
        for (int i=0; i+inc<(*cur).mControl.size(); i+=inc)
          for (ScalarType t=0; t<1.0; t+=1./16) //t+=2.) //t+=1./16.)
            {
              // evaluate Bezier (*cur).mControl[i..i+3] 
              DomainType curpos;
              if (is_cubic)
                {
                  DomainType temp[5];
                  //temp[0] = (1-t)*(*cur).mControl[i];
                  Matrix::EqScale(temp[0], 1-t, (*cur).mControl[i]);
                  //temp[0] += t*(*cur).mControl[i+1];
                  Matrix::AddScaled(temp[0], t, (*cur).mControl[i+1]);
                  //temp[1] = (1-t)*(*cur).mControl[i+1];
                  Matrix::EqScale(temp[1], 1-t, (*cur).mControl[i+1]);
                  //temp[1] += t*(*cur).mControl[i+2];
                  Matrix::AddScaled(temp[1], t, (*cur).mControl[i+2]);
                  //temp[2] = (1-t)*(*cur).mControl[i+2];
                  Matrix::EqScale(temp[2], 1-t, (*cur).mControl[i+2]);
                  //temp[2] += t*(*cur).mControl[i+3];
                  Matrix::AddScaled(temp[2], t, (*cur).mControl[i+3]);
                  //temp[3] = (1-t)*temp[0];
                  Matrix::EqScale(temp[3], 1-t, temp[0]);
                  //temp[3] += t*temp[1];
                  Matrix::AddScaled(temp[3], t, temp[1]);
                  //temp[4] = (1-t)*temp[1];
                  Matrix::EqScale(temp[4], 1-t, temp[1]);
                  //temp[4] += t*temp[2];
                  Matrix::AddScaled(temp[4], t, temp[2]);
                  //curpos = (1-t)*temp[3];
                  Matrix::EqScale(curpos, 1-t, temp[3]);
                  //curpos += t*temp[4];
                  Matrix::AddScaled(curpos, t, temp[4]);
                }
              else
                {
                  Matrix::EqScale(curpos, 1-t, (*cur).mControl[i]);
                  Matrix::AddScaled(curpos, t, (*cur).mControl[i+1]);
                }

              ScalarType x=curpos(0,0), y=curpos(0,1);
              ScalarType value;
              ScalarType gradX;
              ScalarType gradY;
              cubic->ValueAndGrad(x, y, value, gradX, gradY);
              value = Matrix::Abs(value);
              ScalarType grad = sqrt(gradX*gradX+gradY*gradY);
              ScalarType error = 0;
              if (grad>0)
                error = value/grad;
              ScalarType weight = Matrix::Length(curpos, lastpos);
              totalValue += value*weight;
              totalError += error*weight;
              totalWeight += weight;
              if (value>ret->mMaxValue.mValue)
                {
                  ret->mMaxValue.mError=error;
                  ret->mMaxValue.mValue=value;
                  ret->mMaxValue.mGrad=grad;
                  ret->mMaxValue.mX=x;
                  ret->mMaxValue.mY=y;
                }
              if (error>ret->mMaxError.mError)
                {
                  ret->mMaxError.mError=error;
                  ret->mMaxError.mValue=value;
                  ret->mMaxError.mGrad=grad;
                  ret->mMaxError.mX=x;
                  ret->mMaxError.mY=y;
                }
            }
      }
    ret->mAvgValue = totalValue/totalWeight;
    ret->mAvgError = totalError/totalWeight;
  }

ScalarType ComputeError(FunctionCubic2D *cubic, ContourList &results, 
                        bool is_cubic)
  {
    ErrorData errordata;
    ComputeErrorSilent(cubic, results, is_cubic, &errordata);
    printf("Average |value| = %g, average error = %g\n", 
           errordata.mAvgValue, errordata.mAvgError);
    printf("Max error = %g at |value| = %g, |grad| = %g, (%g,%g)\n", 
           errordata.mMaxError.mError, errordata.mMaxError.mValue, 
           errordata.mMaxError.mGrad, errordata.mMaxError.mX, 
           errordata.mMaxError.mY);
    printf("Max |value| = %g at error = %g, |grad| = %g, (%g,%g)\n", 
           errordata.mMaxValue.mValue, errordata.mMaxValue.mError, 
           errordata.mMaxValue.mGrad, errordata.mMaxValue.mX, 
           errordata.mMaxValue.mY);
    return errordata.mMaxError.mError;
  }

template<class Contour_T>
bool ContourOnce(Contour_T &contour, const char *filename, bool is_cubic)
  {
    typedef Contour_T ContourType;
    typedef ContourType::ContourList Results;
    typedef Results::iterator Iter;
    printf("Contouring with epsilon=%g:\n", contour.mEpsilon);

    FunctionProfiled *func=static_cast<FunctionProfiled *>
      (contour.mInterpolant.mFunction);
    func->ResetCounts();

    contour.Init();
    bool workleft;
    do
      {
        //contour.mMesh.Visualize();
        workleft = contour.Step();
      } while (workleft);
    printf("%d elements to be contoured by the interpolant.\n", 
           contour.mMesh.ResultSetSize());
    Results results;
    contour.Results(results);
    if (results.empty())
      printf("-- no contours found --\n");
    else
      PrintContours(results);
    printf("\n");

    printf("Function evaluations: %d\nGrad evals: %d\n",
            func->mFCount, func->mGCount);

    ScalarType errorseen=ComputeError(func, results, is_cubic);

    EpsContours(contour.mInterpolant.mFunction, results, filename, is_cubic);
    return errorseen <= epsilon;
  }

static ScalarType sOptimism = 0.4;
static ScalarType sErrorBound = 1;
static ScalarType sThirdOrder = 100;
static ScalarType sFifthOrder = 1;

static bool Contour2DTestMaxD(FunctionProfiled &cubic, const char *fstr, const char *fn)
  {
    ContourTypeMaxD contour;
    contour.mInterpolant.mFunction = &cubic;
    contour.mInterpolant.mOptimism = sOptimism;
    // bound on the third derivative of the interpolation error
    contour.mInterpolant.mErrorBound = sErrorBound; 

    ContourTypeMaxD::MeshType::PositionType ll_p, ul_p, ur_p;
    ll_p(0,0)=0;
    ll_p(0,1)=0;
    ul_p(0,0)=0;
    ul_p(0,1)=1;
    ur_p(0,0)=1;
    ur_p(0,1)=1;
        // forms diamond ll, ul, ur, ll+ur-ur==lr
    contour.mMesh.InitializeDiamond(ll_p, ul_p, ur_p);

    contour.mEpsilon = epsilon;
    printf(fstr, "MaxD");
    bool ret=ContourOnce(contour, fn, true);

    return ret;
  }

static bool Contour2DTestCubic(FunctionProfiled &cubic, const char *fstr, const char *fn)
  {
    ContourTypeReg contour;
    contour.mInterpolant.mFunction = &cubic;
    contour.mInterpolant.mErr.mOptimism = sOptimism;
    // safety factor
    contour.mInterpolant.mErr.mConservativeError=2;
    contour.mInterpolant.mErr.mFifthOrderError=sFifthOrder;

    ContourTypeReg::MeshType::PositionType ll_p, ul_p, ur_p;
    ll_p(0,0)=0;
    ll_p(0,1)=0;
    ul_p(0,0)=0;
    ul_p(0,1)=1;
    ur_p(0,0)=1;
    ur_p(0,1)=1;
        // forms diamond ll, ul, ur, ll+ur-ur==lr
    contour.mMesh.InitializeDiamond(ll_p, ul_p, ur_p);

    contour.mEpsilon = epsilon;

    printf(fstr, "Cubic");
    bool ret=ContourOnce(contour, fn, true);

    return ret;
  }

static bool Contour2DTestPrecise(FunctionProfiled &cubic, const char *fstr, const char *fn)
  {
    ContourTypePr contour;
    contour.mInterpolant.mFunction = &cubic;
    contour.mInterpolant.mErr.mOptimism = sOptimism;
    // safety factor
    contour.mInterpolant.mErr.mConservativeError=2;
    contour.mInterpolant.mErr.mFifthOrderError=sFifthOrder;

    ContourTypeReg::MeshType::PositionType ll_p, ul_p, ur_p;
    ll_p(0,0)=0;
    ll_p(0,1)=0;
    ul_p(0,0)=0;
    ul_p(0,1)=1;
    ur_p(0,0)=1;
    ur_p(0,1)=1;
        // forms diamond ll, ul, ur, ll+ur-ur==lr
    contour.mMesh.InitializeDiamond(ll_p, ul_p, ur_p);

    contour.mEpsilon = epsilon;
    //contour.mEpsilonRatio = 0.98;

    printf(fstr, "Precise");
    bool ret=ContourOnce(contour, fn, true);

    return ret;
  }

static bool Contour2DTestLinear(FunctionProfiled &cubic, const char *fstr, const char *fn)
  {
    ContourTypeLinear contour;
    contour.mInterpolant.mFunction = &cubic;
    // safety factor
    contour.mInterpolant.mConservativeError=3;
    contour.mInterpolant.mThirdOrderError=sThirdOrder;

    ContourTypeLinear::MeshType::PositionType ll_p, ul_p, ur_p;
    ll_p(0,0)=0;
    ll_p(0,1)=0;
    ul_p(0,0)=0;
    ul_p(0,1)=1;
    ur_p(0,0)=1;
    ur_p(0,1)=1;
        // forms diamond ll, ul, ur, ll+ur-ur==lr
    contour.mMesh.InitializeDiamond(ll_p, ul_p, ur_p);

    contour.mEpsilon = epsilon;

    printf(fstr, "Linear");
    bool ret=ContourOnce(contour, fn, false);

    return ret;
  }

static void InitCubic(CubicToContour &cubic)
  {
    cubic.mCubic.mY[9]=  58.5;
    cubic.mCubic.mY[7]=  8.5;   cubic.mCubic.mY[8]= -49./6;
    cubic.mCubic.mY[4]= -49./6; cubic.mCubic.mY[5]= -149./6;  
                    cubic.mCubic.mY[6]= -49./6;
    cubic.mCubic.mY[0]=  8.5;   cubic.mCubic.mY[1]= -49./6;   
                    cubic.mCubic.mY[2]=  8.5;   cubic.mCubic.mY[3] = 58.5;
    cubic.Init();
  }


static void InitBiCubic1(BiCubicToContour &bicubic, 
                         ScalarType offset)
  { // [y][x]
    bicubic.mCoeff[0][0]=  81;
    bicubic.mCoeff[0][1]= -117;
    bicubic.mCoeff[0][2]=  117;
    bicubic.mCoeff[0][3]= -81;

    bicubic.mCoeff[1][0]= -117;
    bicubic.mCoeff[1][1]=  169;
    bicubic.mCoeff[1][2]= -169;
    bicubic.mCoeff[1][3]=  117;

    bicubic.mCoeff[2][0]=  117;
    bicubic.mCoeff[2][1]= -169;
    bicubic.mCoeff[2][2]=  169;
    bicubic.mCoeff[2][3]= -117;

    bicubic.mCoeff[3][0]= -81;
    bicubic.mCoeff[3][1]=  117;
    bicubic.mCoeff[3][2]= -117;
    bicubic.mCoeff[3][3]=  81;

    for (int y=0; y<4; ++y)
      for (int x=0; x<4; ++x)
        bicubic.mCoeff[y][x]+=offset;

    bicubic.Init();
  }

static void InitBiCubic2(BiCubicToContour &bicubic, 
                         ScalarType offset)
  { // [y][x]
    bicubic.mCoeff[0][0]=    0;
    bicubic.mCoeff[0][1]=   51;
    bicubic.mCoeff[0][2]= - 48;
    bicubic.mCoeff[0][3]=  828;

    bicubic.mCoeff[1][0]= - 51;
    bicubic.mCoeff[1][1]=    0;
    bicubic.mCoeff[1][2]= -374;
    bicubic.mCoeff[1][3]= - 48;

    bicubic.mCoeff[2][0]=   48;
    bicubic.mCoeff[2][1]=  374;
    bicubic.mCoeff[2][2]=    0;
    bicubic.mCoeff[2][3]=   51;

    bicubic.mCoeff[3][0]= -828;
    bicubic.mCoeff[3][1]=   48;
    bicubic.mCoeff[3][2]= - 51;
    bicubic.mCoeff[3][3]=    0;

    for (int y=0; y<4; ++y)
      for (int x=0; x<4; ++x)
        bicubic.mCoeff[y][x]+=offset;

    bicubic.Init();
  }


bool Contour2DTest()
  {
    bool good=true;
/*
    {
      CubicToContour cubic;
      InitCubic(cubic);
      const char *fstr="** Circle ** %1 (10*x-5/2)^2+(10*y-5/2)^2-4\n";
      epsilon = 0.0625;
      good = Contour2DTestMaxD(cubic, fstr, "circlemd.eps");
      good = Contour2DTestCubic(cubic, fstr, "circlecb.eps") && good;
      good = Contour2DTestLinear(cubic, fstr, "circleln.eps") && good;
    }
*/
    {
      BiCubicToContour bicubic;
      InitBiCubic1(bicubic, 0.0125);
      const char *fstr="** Bicubic1 ** %s\n";

      epsilon = 0.2;
      sOptimism = 0.3;
      sErrorBound = 47134;
      good = Contour2DTestMaxD(bicubic, fstr, "bicubicmd.eps") && good;

      epsilon = 0.2;
      sFifthOrder = 1.e8;
      sOptimism = 0.3;
      good = Contour2DTestCubic(bicubic, fstr, "bicubiccb.eps") && good;

      sFifthOrder = 1.e8;
      epsilon = 0.2;
      good = Contour2DTestPrecise(bicubic, fstr, "bicubicpr.eps") && good;

      epsilon = 0.03;
      sThirdOrder=2000;
      good = Contour2DTestLinear(bicubic, fstr, "bicubicln.eps") && good;
    }
/*
    {
      BiCubicToContour bicubic;
      InitBiCubic2(bicubic, 0.01);
      const char *fstr="** Bicubic2 ** %s\n";

      epsilon = 0.05;
      sOptimism = 0.3;
      sErrorBound = 100000;
      good = Contour2DTestMaxD(bicubic, fstr, "thetamd.eps") && good;

      epsilon = 0.05;
      sFifthOrder = 0;
      good = Contour2DTestCubic(bicubic, fstr, "thetacb.eps") && good;

      epsilon = 0.01;
      sOptimism = 0.3;
      sFifthOrder = 0;
      good = Contour2DTestPrecise(bicubic, fstr, "thetapr.eps") && good;

      epsilon = 0.05;
      sThirdOrder=8000;
      good = Contour2DTestLinear(bicubic, fstr, "thetaln.eps") && good;
    }
*/
    return good;
  }

#if defined(PROFILE_HOOKS) && defined(PROFILE_2D)

struct ProfileData
  {
    ScalarType mMaxError;
    ScalarType mAvgError;
    int mElementsContoured;
    int mOutputVerts;
    int mContourSteps;
    int mFCount;
    int mGCount;
  };

template<class Contour_T>
void ProfileContourTemplate(Contour_T &contour, bool is_cubic, ProfileData *profiledata)
  {
    typedef Contour_T ContourType;
    typedef ContourType::ContourList Results;
    typedef Results::iterator Iter;

    ContourTypeMaxD::MeshType::PositionType ll_p, ul_p, ur_p;
    ll_p(0,0)=0;
    ll_p(0,1)=0;
    ul_p(0,0)=0;
    ul_p(0,1)=1;
    ur_p(0,0)=1;
    ur_p(0,1)=1;
        // forms diamond ll, ul, ur, ll+ur-ur==lr
    contour.mMesh.InitializeDiamond(ll_p, ul_p, ur_p);

    contour.mEpsilon = epsilon;

    FunctionProfiled *func=static_cast<FunctionProfiled *>
                            (contour.mInterpolant.mFunction);

    func->ResetCounts();
    
    contour.Init();

    profiledata->mContourSteps=0;
    Results results;
    {
      Profile_Scope(contour_once);
      {
        Profile_Scope(mesh_refinement);

        bool workleft;
        do
          {
            workleft = contour.Step();
            ++profiledata->mContourSteps;
          } while (workleft);
      }

      {
        profiledata->mElementsContoured=contour.mMesh.ResultSetSize();
        {
          Profile_Scope(process_results);
          contour.Results(results);
        }
      }
    }

    profiledata->mFCount=func->mFCount;
    profiledata->mGCount=func->mGCount;

    profiledata->mOutputVerts=0;
    for (Iter res_iter=results.begin(); res_iter!=results.end(); ++res_iter)
      profiledata->mOutputVerts += (*res_iter).mControl.size();

    {
      ErrorData errordata;
      ComputeErrorSilent(func, results, is_cubic, &errordata);
      profiledata->mAvgError = errordata.mAvgError;
      profiledata->mMaxError = errordata.mMaxError.mError;
    }
  }

typedef void (*ProfileFunc)(FunctionProfiled &, ProfileData *);

static void Profile2DMaxD(FunctionProfiled &f, ProfileData *profiledata)
  {
    typedef ContourTypeMaxD ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;

    contour.mInterpolant.mOptimism = sOptimism;
    contour.mInterpolant.mErrorBound = sErrorBound;

    ProfileContourTemplate(contour, true, profiledata);
  }

static void Profile2DCubic(FunctionProfiled &f, ProfileData *profiledata)
  {
    typedef ContourTypeReg ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;

    contour.mInterpolant.mErr.mOptimism = sOptimism;
    contour.mInterpolant.mErr.mConservativeError=2;
    contour.mInterpolant.mErr.mFifthOrderError=sFifthOrder;

    ProfileContourTemplate(contour, true, profiledata);
  }

static void Profile2DPrecise(FunctionProfiled &f, ProfileData *profiledata)
  {
    typedef ContourTypePr ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;

    contour.mInterpolant.mErr.mOptimism = sOptimism;
    contour.mInterpolant.mErr.mConservativeError=2;
    contour.mInterpolant.mErr.mFifthOrderError=sFifthOrder;
    //contour.mEpsilonRatio = 0.98;

    ProfileContourTemplate(contour, true, profiledata);
  }

static void Profile2DLinear(FunctionProfiled &f, ProfileData *profiledata)
  {
    typedef ContourTypeLinear ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;

    contour.mInterpolant.mConservativeError=3;
    contour.mInterpolant.mThirdOrderError=sThirdOrder;

    ProfileContourTemplate(contour, false, profiledata);
  }

static ProfileFunc sMethodsToProfile[] =
  {
    Profile2DMaxD, Profile2DCubic, Profile2DPrecise, Profile2DLinear
  };

const int cNumMethods=sizeof(sMethodsToProfile)/sizeof(sMethodsToProfile[0]);

static const char * sMethodNames[cNumMethods] =
  {
    "MaxD", "Cubic", "Precise", "Linear"
  };

static void Profile2D(FunctionProfiled &f, Profiler &profiler)
  {
    for (int method=0; method<cNumMethods; ++method)
      {
        printf("#%s Method ----------------------\n", sMethodNames[method]);
        printf("# Eps, #F, #G, MaxErr, AvgErr, #Contour, #Elts, #Verts, TimeProfile: name=(#Exec, whole, self)\n");
        for (epsilon=0.125; epsilon>0.00000005; epsilon*=.5)
          {
            ProfileData profiledata;
            profiler.Start();
            sMethodsToProfile[method](f, & profiledata);
            profiler.Done();
            // print report
            printf("%g, %d, %d, %g, %g, %d, %d, %d", 
                    epsilon, profiledata.mFCount, 
                    profiledata.mGCount,
                    profiledata.mMaxError, profiledata.mAvgError,
                    profiledata.mContourSteps,
                    profiledata.mElementsContoured,
                    profiledata.mOutputVerts);
            double maxwholetime=0;
            for (Profiler::ReportIter report_item=profiler.mReport.begin();
                  report_item!=profiler.mReport.end(); ++report_item)
              {
                printf(", %s=(%d, %f, %f)",
                  (*report_item).mName, (*report_item).mNumExecutions, 
                  (*report_item).mWholeTime, (*report_item).mSelfTime);
                if ((*report_item).mWholeTime>maxwholetime)
                  maxwholetime=(*report_item).mWholeTime;
              }
            printf("\n");
            fflush(stdout);
            if (maxwholetime>=1000)
              break;
          }
        printf("\n");
      }
  }

void ProfileContour2D()
  {
    std::auto_ptr<Profiler> profiler(Profiler::Create());
    {
      CubicToContour cubic;
      InitCubic(cubic);
      printf("#******************************** circle\n");

      sOptimism = 0.4;
      sErrorBound = 0;
      sThirdOrder = 0;
      sFifthOrder = 0;

      Profile2D(cubic, *profiler);
    }
    {
      BiCubicToContour bicubic;
      InitBiCubic1(bicubic, 0.0125);
      printf("#******************************** bicubic1\n");

      sOptimism = 0.3;
      sErrorBound = 47134;
      sFifthOrder = 1.e8;
      sThirdOrder=2000;
      Profile2D(bicubic, *profiler);
    }
    {
      BiCubicToContour bicubic;
      InitBiCubic2(bicubic, 0.01);
      printf("#******************************** theta\n");

      sOptimism = 0.3;
      sErrorBound = 100000;
      sFifthOrder = 0;
      sThirdOrder=8000;
      Profile2D(bicubic, *profiler);
    }
  }

#endif