#include "int2dln.h"
#include "matrix.h"
#include "puzzle.h"
#include "merge2d.h"
#include "int2dbs.h"
#include "blow_profile/profiler.h"

// begin debug code
#if ! defined(NDEBUG) && ! defined(PROFILE_HOOKS)
#define INT_2D_VIS
#include "draw.h"
#endif
// end debug code


Extern_Zone(contour_interpolant);
Extern_Zone(puzzle_piece);
Extern_Zone(puzzle_finish);

typedef InterpolantLinear2D::VertexType VertexType;
typedef std::list<InterpolantLinear2D::ContourType> ContourList;

static ScalarType Gradient(VertexType *v0, VertexType *v1, VertexType *v2, 
                           ScalarType radius2)
  {
    assert(radius2>0);
    ScalarType d1 = (v0->mFunctionValue-v1->mFunctionValue)/radius2;
    ScalarType d2 = (v2->mFunctionValue-v1->mFunctionValue)/radius2;
    return sqrt(d1*d1+d2*d2);
  }

#define SQRT2 (1.4142135623730950488016887242097)

ElementAction InterpolantLinear2D 
::AnalyzeElement(MeshElement_VirtualBase<VertexType> *element, 
                                 ScalarType radius, VertexType *center, 
                                 ScalarType epsilon)
  {
    int i;
    int numcorners=element->NumCorners();
    assert(numcorners==3 || numcorners==4);
    VertexType *v[4];

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated & VertexType::ValueEvaluated);
      }

    ScalarType val;
    if (numcorners==3)
      val=(v[0]->mFunctionValue+v[2]->mFunctionValue)*.5;
    else
      val=(v[0]->mFunctionValue+v[1]->mFunctionValue
          +v[2]->mFunctionValue+v[3]->mFunctionValue)*.25;
    ScalarType error=Matrix::Abs(center->mFunctionValue-val);
    error+=mThirdOrderError*radius*radius*radius;
    error*=mConservativeError;

    if (v[0]->mFunctionValue>0)
      {
        if (center->mFunctionValue-error>0)
          {
            for (i=0; i<numcorners; ++i)
              if (v[i]->mFunctionValue-error<=0)
                break;
            if (i>=numcorners)
              return DiscardElement;
          }
      }
    else
      {
        if (center->mFunctionValue+error<0)
          {
            for (i=0; i<numcorners; ++i)
              if (v[i]->mFunctionValue+error>=0)
                break;
            if (i>=numcorners)
              return DiscardElement;
          }
      }
    
    ScalarType min_grad;
    /* compute minimum derivative */
    if (numcorners==3)
      {
        min_grad = Matrix::Abs(Gradient(v[0], v[1], v[1], radius*SQRT2));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[1], center, v[0], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[2], center, v[1], radius)));
      }
    else
      {
        VertexType interp(*center);
        interp.mFunctionValue=val;
        min_grad = Matrix::Abs(Gradient(v[0], center, v[3], radius));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[1], center, v[0], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[2], center, v[1], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[3], center, v[2], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[0], &interp, v[3], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[1], &interp, v[0], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[2], &interp, v[1], radius)));
        min_grad = Matrix::Min(min_grad, 
                   Matrix::Abs(Gradient(v[3], &interp, v[2], radius)));
      }

    if (error < epsilon*min_grad)
      return ContourElement;
    return RefineElement;
  }

static void LinearContour2d(ContourList &results, VertexType *v[3], ScalarType epsilon)
  {
    Profile_Scope(contour_interpolant);
    assert(epsilon>=0);
    int i;
    bool signchange[3];
    bool cornerzero[3];
    ScalarType intersect[3];
    bool output=false;

    for (i=0; i<3; ++i)
      cornerzero[i]=false;
    for (i=0; i<3; ++i)
      {
        ScalarType y[2];
        y[0]=v[i]->mFunctionValue;
        y[1]=v[(i+1)%3]->mFunctionValue;
        signchange[i]=(Root1D::SignOfProduct(y[0], y[1])==Root1D::Negative);
        if (signchange[i])
          {
            output=true;
            intersect[i]=Matrix::Abs(y[0]/(y[0]-y[1]));
            if (intersect[i]<epsilon)
              {
                signchange[i]=false;
                cornerzero[i]=true;
              }
            else if (intersect[i]>1-epsilon)
              {
                signchange[i]=false;
                cornerzero[(i+1)%3]=true;
              }
          }
      }
    if (output)
      {
        int side[2];
        DomainType2d out[2];
        int numout=0;
        for (i=0; i<3; ++i)
          if (signchange[i])
            {
              assert(numout<2);
              out[numout]=v[i]->mPos;
              Matrix::AddScaledDiff(out[numout], intersect[i], 
                                    v[(i+1)%3]->mPos, v[i]->mPos);
              side[numout]=i;
              ++numout;
            }
        for (i=0; i<3; ++i)
          if (cornerzero[i])
            {
              if (numout==2)
                {
                  results.push_back();
                  results.back().mEndingSides[0]=side[0];
                  results.back().mEndingSides[1]=side[1];
                  results.back().mControl.push_back(out[0]);
                  results.back().mControl.push_back(out[1]);
                  numout=0;
                }
              assert(numout<2);
              out[numout]=v[i]->mPos;
              side[numout]=i+3;
              ++numout;
            }
        if (numout)
          {
            results.push_back();
            results.back().mEndingSides[0]=side[0];
            results.back().mControl.push_back(out[0]);
            if (numout>1)
              {
                results.back().mEndingSides[1]=side[1];
                results.back().mControl.push_back(out[1]);
              }
            else
              {
                assert(side[0]>=3);
                results.back().mControl.push_back(out[0]);
                results.back().mEndingSides[1]=side[0];
              }
          }
      }
  }

void InterpolantLinear2D 
::InterpolateContour(std::list<ContourType> &results, 
                     MeshElementList_VirtualBase<VertexType> *elt_lst)
  {
    MeshElement_VirtualBase<VertexType> *element;
    Puzzle puzzle(mEpsilon);

        // BEGIN DEBUG
#ifdef INT_2D_VIS
    static char filename[10];
    itoa(gDebugOutputFileCount, filename, 10);
    strcat(filename, ".eps");
    ++gDebugOutputFileCount;
    PS::Drawing drawing;
    PS::State state=drawing.GetState();
    PS::State contourstate(state);
    state.SetLineThickness(state.GetLineThickness()*.5);
    {
      PS::Path path;
      path.MoveTo(ScaleUp(0), ScaleUp(0));
      path.LineTo(ScaleUp(1), ScaleUp(0));
      path.LineTo(ScaleUp(1), ScaleUp(1));
      path.LineTo(ScaleUp(0), ScaleUp(1));
      path.Close();
      state.SetGray(0.9);
      drawing.StrokePath(path, state);
    }
#endif
        // END DEBUG

    while ((element=elt_lst->NextElement())!=0)
      {
        assert(element->NumCorners()==3 || element->NumCorners()==4);
        VertexType *v[4];
        int i;
        int numcorners=element->NumCorners();

        for (i=0; i<numcorners; ++i)
          {
            v[i]=& element->Corner(i);
            if (!(v[i]->mEvaluated & VertexType::ValueEvaluated))
              Evaluate(*v[i]);
            assert(v[i]->mEvaluated == VertexType::BothEvaluated);
          }

        // BEGIN DEBUG
#ifdef INT_2D_VIS
        {
          PS::Path path;
          path.MoveTo(ScaleUp(v[0]->mPos(0,0)), ScaleUp(v[0]->mPos(0,1)));
          for (i=1; i<numcorners; ++i)
            path.LineTo(ScaleUp(v[i]->mPos(0,0)), ScaleUp(v[i]->mPos(0,1)));
          path.Close();
          state.SetGray(.5);
          drawing.StrokePath(path, state);
        }
#endif
        // END DEBUG

        if (numcorners==3)
          {
            ContourList contour;
            LinearContour2d(contour, v, mEpsilon);
            int neighbors[3];
            DomainType2d cornerpos[3];
            for (int side=0; side<3; ++side)
              {
                neighbors[side]=element->Neighbor(side);
                cornerpos[side]=v[side]->mPos;
              }
            {
              Profile_Scope(puzzle_piece);
              puzzle.AddPiece(results, element->Id(), 3, neighbors, contour, cornerpos);
            }
          }
        else
          {
            // puzzle.AddPiece(results, element->Id(), 4, neighbors, contours[0]);
            VertexType center(v[0]->mPos);
            Matrix::Add(center.mPos, center.mPos, v[2]->mPos);
            Matrix::Scale(center.mPos, 0.5);
            center.mFunctionValue=(v[0]->mFunctionValue+v[1]->mFunctionValue
                                  +v[2]->mFunctionValue+v[3]->mFunctionValue)*.25;

            static int renum[4][6][2] = 
              { { { 0,  8 }, { 1, 0 }, { 2,  9 }, { 3, 12 }, { 4, 4 }, { 5, 5 } },
                { { 0,  9 }, { 1, 1 }, { 2, 10 }, { 3, 12 }, { 4, 5 }, { 5, 6 } },
                { { 0, 10 }, { 1, 2 }, { 2, 11 }, { 3, 12 }, { 4, 6 }, { 5, 7 } },
                { { 0, 11 }, { 1, 3 }, { 2,  8 }, { 3, 12 }, { 4, 7 }, { 5, 4 } }
              };
            static int tomerge1[1] = { 9 };
            static int tomerge2[1] = { 10 };
            static int tomerge3[3] = { 8, 11, 12 };

            int side;
            ContourList contours[4];
            for (side=0; side<4; ++side)
              {
                VertexType *verts[3];
                verts[0]=&center;
                verts[1]=v[side];
                verts[2]=v[(side+1)%4];
                LinearContour2d(contours[side], verts, mEpsilon);
                Interp2D::RenumberContours(contours[side], 6, renum[side]);
              }
            {
              Profile_Scope(puzzle_piece);
              Merge2D::CornerNbhd corners[3];
              corners[0].mSide=12;
              corners[0].mLocation=center.mPos;
              corners[1].mSide=5;
              corners[1].mLocation=v[1]->mPos;
              Merge2D::MergeContours(contours[0], contours[1], mEpsilon, 1, tomerge1, 2, corners);
              corners[1].mSide=6;
              corners[1].mLocation=v[2]->mPos;
              Merge2D::MergeContours(contours[0], contours[2], mEpsilon, 1, tomerge2, 2, corners);
              corners[1].mSide=7;
              corners[1].mLocation=v[3]->mPos;
              corners[2].mSide=4;
              corners[2].mLocation=v[0]->mPos;
              Merge2D::MergeContours(contours[0], contours[3], mEpsilon, 3, tomerge3, 3, corners);
              assert(contours[1].empty());
              assert(contours[2].empty());
              assert(contours[3].empty());
              int neighbors[4];
              DomainType2d cornerpos[4];
              for (side=0; side<4; ++side)
                {
                  neighbors[side]=element->Neighbor(side);
                  cornerpos[side]=v[side]->mPos;
                }
              puzzle.AddPiece(results, element->Id(), 4, neighbors, contours[0], cornerpos);
            }
          }
      }
    {
      Profile_Scope(puzzle_finish);
      puzzle.FinishPuzzle(results);
    }
        // BEGIN DEBUG
#ifdef INT_2D_VIS
    DrawLinearContour(results, drawing, contourstate);
    // draw ideal
    //state.SetRGB(0,1,0);
    //  {
    //    PS::Path path;
    //    path.Circle(ScaleUp(.25), ScaleUp(.25), Radius(.2));
    //    drawing.StrokePath(path, state);
    //  }
    WriteEPSFile(drawing, filename, "InterpolateContour", "Shape");
#endif
        // END DEBUG
  }
