#include "int2dpr.h"

static void ConstructCubicTriangle(Root2D::TriangularCubic &cubic, 
                            Interpolant2DPrecise::FunctionType *func,
                            Vertex_2D_Cubic_Base *v0,
                            Vertex_2D_Cubic_Base *v1,
                            DomainType2d center)
  {
    using namespace Matrix;

    cubic.mP[0]=center;
    cubic.mP[1]=v0->mPos;
    cubic.mP[2]=v1->mPos;

    cubic.mY[3]=v0->mFunctionValue;
    cubic.mY[9]=v1->mFunctionValue;

    DomainType2d midpoint;
    Add(midpoint, v0->mPos, v1->mPos);
    Scale(midpoint, .5);

    DomainType2d edges[3];
    int i;
    for (i=0; i<3; ++i)
      Subtract(edges[i], cubic.mP[(i+1)%3], cubic.mP[i]);
    cubic.mY[2]=cubic.mY[3]-Dot(edges[0], v0->mGradient)*(1./3);
    cubic.mY[6]=cubic.mY[3]+Dot(edges[1], v0->mGradient)*(1./3);
    cubic.mY[8]=cubic.mY[9]-Dot(edges[1], v1->mGradient)*(1./3);
    cubic.mY[7]=cubic.mY[9]+Dot(edges[2], v1->mGradient)*(1./3);

    DomainType2d perp;
    Subtract(perp, center, midpoint);
    ScalarType crossboundaryderiv = func->PartialDeriv(midpoint(0,0),
      midpoint(0,1), perp(0,0), perp(0,1));
    cubic.mY[5]=(2./3)*crossboundaryderiv + .25*
      (3*(cubic.mY[6]+cubic.mY[8])-2*(cubic.mY[7]+cubic.mY[2])
       + cubic.mY[3] + cubic.mY[9]);
  }


void PreciseInterpStrategy::TriangleInterpolant(PreciseBase *interp,
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR)
  {
    Interpolant2DPrecise::FunctionType *func=
      static_cast<Interpolant2DPrecise *>(interp)->mFunction;

    DomainType2d center=v[0]->mPos;
    int side;
    for (side=1; side<3; ++side)
      Matrix::Add(center, center, v[side]->mPos);
    Matrix::Scale(center, 1./3);

    ConstructCubicTriangle(cubics[0], func, v[0], v[1], center);
    ConstructCubicTriangle(cubics[1], func, v[1], v[2], center);
    ConstructCubicTriangle(cubics[2], func, v[2], v[0], center);

    // figure out cubics[side].mY[1,4]
    for (side=0; side<3; ++side)
      {
        int otherside=(side+2)%3;
        assert(Matrix::Abs(cubics[side].mY[2]-cubics[otherside].mY[7])
                  <epsilonR);
        assert(Matrix::Abs(cubics[side].mY[3]-cubics[otherside].mY[9])
                  <epsilonR);
        // this may only be right for equalateral triangles!
        // but it might actually be correct in general, weird
        cubics[side].mY[1] = (cubics[side].mY[2]+cubics[side].mY[5]+
                              cubics[otherside].mY[5])*(1./3);
        cubics[otherside].mY[4] = cubics[side].mY[1];
      }

    // figure out cubics[side].mY[0]
    cubics[0].mY[0] = (cubics[0].mY[1]+cubics[0].mY[4]+cubics[1].mY[4])
                        *(1./3);
    cubics[1].mY[0] = cubics[0].mY[0];
    cubics[2].mY[0] = cubics[0].mY[0];
  }

void PreciseInterpStrategy::SquareInterpolant(PreciseBase *interp,
                        Vertex_2D_Cubic_Base **v,
                        Root2D::TriangularCubic *cubics,
                        ScalarType epsilonR)
  {
    Interpolant2DPrecise::FunctionType *func=
      static_cast<Interpolant2DPrecise *>(interp)->mFunction;

    DomainType2d center=v[0]->mPos;
    int side;
    for (side=1; side<4; ++side)
      Matrix::Add(center, center, v[side]->mPos);
    Matrix::Scale(center, 1./4);

    ConstructCubicTriangle(cubics[0], func, v[0], v[1], center);
    ConstructCubicTriangle(cubics[1], func, v[1], v[2], center);
    ConstructCubicTriangle(cubics[2], func, v[2], v[3], center);
    ConstructCubicTriangle(cubics[3], func, v[3], v[0], center);

    // figure out cubics[side].mY[1,4]
    for (side=0; side<4; ++side)
      {
        int otherside=(side+3)%4;
        assert(Matrix::Abs(cubics[side].mY[2]-cubics[otherside].mY[7])
                  <epsilonR);
        assert(Matrix::Abs(cubics[side].mY[3]-cubics[otherside].mY[9])
                  <epsilonR);
        // note: fewer vertices than above (C-T/triangular) case
        cubics[side].mY[1] = (cubics[side].mY[5]+
                              cubics[otherside].mY[5])*.5;
        // this way, the four vertices are planar!
        cubics[otherside].mY[4] = cubics[side].mY[1];
      }

    // figure out cubics[side].mY[0]
    cubics[0].mY[0] = (cubics[0].mY[1]+cubics[0].mY[4]+
                       cubics[2].mY[1]+cubics[2].mY[4])*.25;
    cubics[1].mY[0] = cubics[0].mY[0];
    cubics[2].mY[0] = cubics[0].mY[0];
    cubics[3].mY[0] = cubics[0].mY[0];
  }

ElementAction Interpolant2DPrecise::AnalyzeElement
    (MeshElement_VirtualBase<VertexType> *element, 
     ScalarType radius, VertexType *center, 
     ScalarType epsilon)
  {
    VertexType *v[4];
    int numcorners=element->NumCorners();
    int i;
    const ScalarType sqrt2=1.414214; // sqrt(2)

    for (i=0; i<numcorners; ++i)
      {
        v[i]=& element->Corner(i);
        assert(v[i]->mEvaluated == VertexType::BothEvaluated);
      }

    ScalarType error;
    if (numcorners==3)
      {
        DomainType edge, edgeunit, edgenormal;
        Matrix::Subtract(edge, v[2]->mPos, v[0]->mPos);
        ScalarType edgelen=2*radius;
        assert(Matrix::Abs(Matrix::TwoNorm(edge)-edgelen)<0.00001);
        Matrix::EqScale(edgeunit, 1/edgelen, edge);
        edgenormal(0,0)= -edgeunit(0,1);
        edgenormal(0,1)= edgeunit(0,0);
        ScalarType g[2];
        g[0] = Matrix::Dot(edgeunit, v[0]->mGradient);
        g[1] = Matrix::Dot(edgeunit, v[2]->mGradient);
        error = (v[0]->mFunctionValue+v[2]->mFunctionValue)*.5
                  +(g[0]-g[1])*radius*(1./4);
        error = Matrix::Abs(center->mFunctionValue-error);

        assert(error>=0);
        // this is intended to be an over estimate
        //ScalarType dn = sqrt(d1*d1+d2*d2);
        //error=mConservativeError*ErrorOfCubic(val, 2*dn, radius*sqrt2);
        ScalarType radius2=radius*radius;
        error += radius2*radius2*radius*mErr.mFifthOrderError;
        error *= mErr.mConservativeError;
        return StoppingCriteriaHelperTriangle
          (error, epsilon, mErr.mOptimism, element);
      }
    else
      {
        assert(numcorners==4);
        Root2D::TriangularCubic cubics[4];
        Interp2D::SibsonSplitSquare(
          reinterpret_cast<Vertex_2D_Cubic_Base **>(v), cubics, mEpsilonR);
        //cubics[*].mY[0] should be compared to center->mFunctionValue
        error=Matrix::Abs(cubics[0].mY[0] - center->mFunctionValue);
        //ScalarType dn = sqrt(d1*d1+d2*d2);
        //error=mConservativeError*ErrorOfCubic(val, 2*dn, radius*sqrt2);
        ScalarType radius2=radius*radius;
        error += radius2*radius2*radius*mErr.mFifthOrderError;
        error *= mErr.mConservativeError;
        return StoppingCriteriaHelperSquare
          (error, epsilon, mErr.mOptimism, element, cubics);
      }
  }
