#ifndef __BTT_MESH_H
#define __BTT_MESH_H

#include "vert.h"
#include <vector>

/* A leaf-only split-only binary-triangle-tree mesh data structure
   a.k.a. newest-vertex-bisection mesh

    Tricky parts:
      * Elements already placed in the result set may need to be refined
        in order to maintain a conforming mesh.  Therefore elements need
        to be able to quickly tell if they are in the result set.
      * Internally the mesh contains (right-isoceles) triangles, but the
        refinement operation is normally applied to two trianles meeting
        base-to-base, ie. they form a "diamond".  To the outside world,
        it reports an element as the union of the two triangles if they
        form a diamond.  Only one member of the diamond is put into the
        working set, but both are put in the result set.
      * In order to avoid having to remove a random element from the 
        working set, the working set is ordered by the diameter of the
        elements.  PopWorking() returns the largest diameter element.
        This should elimate force-splits of elements in the working set
        (elements in the result set still may need to be force split:
        this is the only way elements in the result set are split;
        elements discarded shouldn't: they are disconnected from the 
        topology).  It should also improve the number of triangles 
        considered as part as a diamond instead of individually.
      * Currently mNumResults counts the number of triangles in the 
        result set, not the number of times PopResult() will return
        an element (since it returns both triangles and diamonds).

    Assumptions:
      * The following functions assume elements are right-isoceles:
          Diameter(...)
          IsCoveredByCorners(...)
          RadiusForCenterToCover(...), along with its helper function:
          RadiusToCoverCorner(...)
      * PopWorking implicitly assumes that elements only force split
        elements with a larger diameter.  That is the whole point of
        maintaining a priority queue for the working set ordered by
        diameter.

    Future:
      * If PopResult() is too slow, decent asymptotic behavoir can be
        achieved using a set or a hash_set to store which elements are
        in the result set.
 */

template<class Traits_T>
class BTTMesh
  {
public:
    typedef Traits_T TraitsType;
    typedef BTTMesh<TraitsType> thistype;
    typedef typename TraitsType::DomainType PositionType;
    typedef TraitsType::VertexType VertexType;
    typedef int ElementType; 

protected:
    enum { Left, Right, Base };
    enum { NotResult = -2 };
    class BTTElement
      {
        friend class BTTMesh<TraitsType>;
        int mVerts[3];
        int mCenter;
        ElementType mNeighbors[3];
        ElementType mNextResult;  // == NotResult if not a member of result set
      };

    typedef std::vector<VertexType> VertPool;
    VertPool mVerts;

    typedef std::vector<BTTElement> ElementPool;
    ElementPool mElements;

    class Compare
      {
        typedef BTTMesh<TraitsType> MeshType;
        MeshType *mMesh;
    public:
        explicit Compare(MeshType *mesh) 
          : mMesh(mesh) { }
        bool operator()(ElementType i, ElementType j)
          {
            ScalarType di = mMesh->Diameter(i);
            ScalarType dj = mMesh->Diameter(j);
            if (di<dj)
              return true;
            if (dj<di)
              return false;
            return i>j; // or i>j, doesn't really matter
          }
      };

    typedef std::vector<ElementType> ListType;  //list, vector, or s_list?
    typedef ListType::iterator    IterType;
    ListType mWorkingSet;
    Compare mCompare;

    void PushToWorkingSet(ElementType elt);
    ElementType PopFromWorkingSet();

    ElementType mFirstResult;
    int mNumResults;

    int mInitial[4];

public:
    static ElementType InvalidElement();
    bool ElementValid(ElementType element);

protected:
    bool IsSameElement(ElementType elt1, ElementType elt2);
    bool InWorking(ElementType element);
    bool InResult(ElementType element);
    bool InWorkingOrResult(ElementType element);
      
public:
    BTTMesh();

        // Initialize ...(...)
        // Creates an initial mesh and inserts it into the working set

        // forms diamond a, b, c, a+c-b
    void InitializeDiamond(PositionType a, PositionType b, PositionType c);

    void InitializeTriangle(PositionType l, PositionType b, PositionType r);

      // Element management functions ------------------------------------
protected:
    bool VertexValid(int vert);

         // Examine left neighbor, reassign its pointer to new left child
         // and if a diamond has been formed, copy center vertex, otherwise
         // need to add this new element to working set
    void DiamondFinder(ElementType elt, ElementType child, ElementType neighbor);

    ElementType SplitHelper(ElementType elt);

public:
        // Refine(...)
        // refine must not be in the working set or result set
    void Refine(ElementType elt);

    int NumCorners(ElementType elt);

protected:
    ElementType Name(ElementType elt);

public:
    int Neighbor(ElementType elt, int which);

    VertexType & Corner(ElementType element, int corner);

    ScalarType Diameter(ElementType element);

    ScalarType Radius(ElementType element);

    bool IsCoveredByCorners(ElementType element, ScalarType elementradius);

    void AllocateCenter(ElementType element);
    
    VertexType & Center(ElementType element);

    ScalarType RadiusForCenterToCover(ElementType element, ScalarType elementradius);

protected:
    ScalarType RadiusToCoverCorner(ScalarType saferadius, ScalarType elementradius);
public:
      // --------------------------------------------------------------

        // PopWorking(...)
        // removes a element from the working set and returns it
        // returns InvalidElement() if working set is empty
        // result should be then either: refined, discarded, or 
        //        added to the results
    ElementType PopWorking();

        // Currently disconnects element from adjacent elements
        // could potentially reduce memory usage here
        // if I feel like visualizing, probably want to keep this data around
    void DiscardElement(ElementType element);

protected:
    void DisconnectFromElement(ElementType element, int edge);

public:
        // PushResult(...)
        // adds a element to the result set
        // must not (already) be part of the result or working set
    void PushResult(ElementType result);

        // PopWorking(...)
        // removes a element from the result set and returns it
        // returns InvalidElement() if result set is empty
    ElementType PopResult();

    int ResultSetSize();

    void Visualize();
  };

#include "meshbtt.hpp"

#endif