#include "contourmd.h"
#include "contour.h"
#include "quadrat.h"
#include "mesh1d.h"
#include "int1dmd.h"
#include "int1d.h"
#include "int1dln.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#if defined(PROFILE_HOOKS) && defined(PROFILE_1D)
#include "blow_profile/profiler.h"
#include "sinxx.h"
Define_Zone(contour_once);
  Define_Zone(mesh_refinement);
  Define_Zone(process_results);
    Define_Zone(interpolate);
    Define_Zone(contour_interpolant);
    Define_Zone(puzzle_piece);
    Define_Zone(puzzle_finish);
static ScalarType epsilon;
#else
static const ScalarType epsilon = 0.00000005;
#endif


typedef QuadraticFunction FunctionType;
typedef InterpolantCubic1DMaxD InterpTypeMaxD;
typedef InterpolantCubic1D InterpTypeReg;
typedef InterpolantLinear1D InterpTypeLinear;
typedef InterpTypeMaxD::TraitsType TraitsMaxD;
typedef InterpTypeReg::TraitsType TraitsReg;
typedef InterpTypeLinear::TraitsType TraitsLinear;
typedef ContourMaxD<LinearMesh<TraitsMaxD>, InterpTypeMaxD> ContourTypeMaxD;
typedef Contour<LinearMesh<TraitsReg>, InterpTypeReg> ContourTypeReg;
typedef Contour<LinearMesh<TraitsLinear>, InterpTypeLinear> ContourTypeLinear;
typedef ContourTypeMaxD::ContourList Results;
typedef Results::iterator Iter;

static bool Near(ScalarType a, ScalarType b)
  {
    ScalarType diff=a-b;
    if (diff<0)
      diff= -diff;
    return diff <= epsilon;
  }

static bool CompareResultLists(const Results & results, 
                           const ScalarType *desired, int num)
  {
    int i;
    Results::const_iterator j;
    for (i=0; i<num; ++i)
      { // find (desired[2i], desired[2i+1]) in results
        for (j=results.begin(); j!=results.end(); ++j)
          {
            if (Near((*j)(0,0),desired[i]))
              break;
          }
        if (j==results.end())
          return false;
      }
    for (j=results.begin(); j!=results.end(); ++j)
      { // find *j in desired
        for (i=0; i<num; ++i)
          {
            if (Near((*j)(0,0),desired[i]))
              break;
          }
        if (i==num)
          return false;
      }
    return true;
  }

bool SolveQuadratic_MaxD(double a, double b, double c, double low, double high, bool verbose)
  {
    double middle = -b/(2*a);
    double determ = b*b-4*a*c;
    if (verbose)
      printf("%g x^2 + %g x + %g = 0; %g <= x <= %g\n", a, b, c, low, high);
    ScalarType expected[2];
    int numexpected=0;
    if (determ>0)
      {
        double delta=sqrt(determ)/(2*a);
        if (verbose)
          printf("Roots: %g, %g\n", middle-delta, middle+delta);
        if (middle-delta>=low && middle-delta<=high)
          expected[numexpected++]=middle-delta;
        if (middle+delta>=low && middle+delta<=high)
          expected[numexpected++]=middle+delta;
      }
    else if (determ<0)
      {
        if (verbose)
          printf("No roots\n");
      }
    else
      {
        if (verbose)
          printf("Double root: %g\n", middle);
        if (middle>=low && middle<=high)
          expected[numexpected++]=middle;
      }

    FunctionType func;
    func.Init(a,b,c);
    ContourTypeMaxD contour;
    contour.mInterpolant.mFunction=&func;
    ContourTypeMaxD::MeshType::PositionType low_p, high_p;
    low_p(0,0)=low;
    high_p(0,0)=high;
    contour.mMesh.Initialize(low_p, high_p);
    contour.mEpsilon = epsilon;
    if (verbose)
      printf("MaxD contouring with epsilon=%g:\n", contour.mEpsilon);
    contour.Init();
    bool workleft;
    do
      {
        if (verbose)
          contour.mMesh.Visualize();
        workleft = contour.Step();
      } while (workleft);
    if (verbose)
      printf("%d elements to be contoured by the interpolant.\n", 
             contour.mMesh.ResultSetSize());
    Results results;
    contour.Results(results);
    if (verbose)
      {
        if (results.empty())
          printf("-- no roots found --\n");
        else
          for (Iter i=results.begin(); i!=results.end(); ++i)
            printf("Root: %g\n", (*i)(0,0));
        printf("\n");
      }
    return CompareResultLists(results, expected, numexpected);
  }

bool SolveQuadratic(double a, double b, double c, double low, double high, bool verbose)
  {
    double middle = -b/(2*a);
    double determ = b*b-4*a*c;
    if (verbose)
      printf("%g x^2 + %g x + %g = 0; %g <= x <= %g\n", a, b, c, low, high);
    ScalarType expected[2];
    int numexpected=0;
    if (determ>0)
      {
        double delta=sqrt(determ)/(2*a);
        if (verbose)
          printf("Roots: %g, %g\n", middle-delta, middle+delta);
        if (middle-delta>=low && middle-delta<=high)
          expected[numexpected++]=middle-delta;
        if (middle+delta>=low && middle+delta<=high)
          expected[numexpected++]=middle+delta;
      }
    else if (determ<0)
      {
        if (verbose)
          printf("No roots\n");
      }
    else
      {
        if (verbose)
          printf("Double root: %g\n", middle);
        if (middle>=low && middle<=high)
          expected[numexpected++]=middle;
      }

    FunctionType func;
    func.Init(a,b,c);
    ContourTypeReg contour;
    contour.mInterpolant.mFunction=&func;
    contour.mInterpolant.mConservativeError=2;

    ContourTypeReg::MeshType::PositionType low_p, high_p;
    low_p(0,0)=low;
    high_p(0,0)=high;
    contour.mMesh.Initialize(low_p, high_p);
    contour.mEpsilon = epsilon;
    if (verbose)
      printf("Regular contouring with epsilon=%g:\n", contour.mEpsilon);
    contour.Init();
    bool workleft;
    do
      {
        if (verbose)
          contour.mMesh.Visualize();
        workleft = contour.Step();
      } while (workleft);
    if (verbose)
      printf("%d elements to be contoured by the interpolant.\n", 
             contour.mMesh.ResultSetSize());
    Results results;
    contour.Results(results);
    if (verbose)
      {
        if (results.empty())
          printf("-- no roots found --\n");
        else
          for (Iter i=results.begin(); i!=results.end(); ++i)
            printf("Root: %g\n", (*i)(0,0));
        printf("\n");
      }
    return CompareResultLists(results, expected, numexpected);
  }

bool SolveQuadratic_Linear(double a, double b, double c, double low, double high, bool verbose)
  {
    double middle = -b/(2*a);
    double determ = b*b-4*a*c;
    if (verbose)
      printf("%g x^2 + %g x + %g = 0; %g <= x <= %g\n", a, b, c, low, high);
    ScalarType expected[2];
    int numexpected=0;
    if (determ>0)
      {
        double delta=sqrt(determ)/(2*a);
        if (verbose)
          printf("Roots: %g, %g\n", middle-delta, middle+delta);
        if (middle-delta>=low && middle-delta<=high)
          expected[numexpected++]=middle-delta;
        if (middle+delta>=low && middle+delta<=high)
          expected[numexpected++]=middle+delta;
      }
    else if (determ<0)
      {
        if (verbose)
          printf("No roots\n");
      }
    else
      {
        if (verbose)
          printf("Double root: %g\n", middle);
        if (middle>=low && middle<=high)
          expected[numexpected++]=middle;
      }

    FunctionType func;
    func.Init(a,b,c);
    ContourTypeLinear contour;
    contour.mInterpolant.mFunction=&func;
    contour.mInterpolant.mConservativeError=3;

    ContourTypeReg::MeshType::PositionType low_p, high_p;
    low_p(0,0)=low;
    high_p(0,0)=high;
    contour.mMesh.Initialize(low_p, high_p);
    contour.mEpsilon = epsilon;
    if (verbose)
      printf("Regular contouring with epsilon=%g:\n", contour.mEpsilon);
    contour.Init();
    bool workleft;
    do
      {
        if (verbose)
          contour.mMesh.Visualize();
        workleft = contour.Step();
      } while (workleft);
    if (verbose)
      printf("%d elements to be contoured by the interpolant.\n", 
             contour.mMesh.ResultSetSize());
    Results results;
    contour.Results(results);
    if (verbose)
      {
        if (results.empty())
          printf("-- no roots found --\n");
        else
          for (Iter i=results.begin(); i!=results.end(); ++i)
            printf("Root: %g\n", (*i)(0,0));
        printf("\n");
      }
    return CompareResultLists(results, expected, numexpected);
  }

static ScalarType frand()
  {
    return ((ScalarType)rand())/(RAND_MAX+1);
  }

typedef bool QuadraticSolver(double a, double b, double c, double low, double high, bool verbose);

static bool PerformContour1DTest(QuadraticSolver quadraticsolver, 
                                 const char *test)
  {
    printf("%s Quadratic/1D Tests\n", test);
    bool good=quadraticsolver(1, 0, -2, 0, 2, true);
    good=quadraticsolver(1, 0, 0, -1./3, 2./3, true) && good;
    good=quadraticsolver(1, 0, -1, -1, 1, true) && good;
    int iterations=1000;
    printf("Testing contouring 1d quadratics (%s) (%d iterations)\n", 
          test, iterations);
    for (int t=0; t<iterations; ++t)
      {
        ScalarType a=frand(), b=frand(), c=frand(), mn=-3*frand(), mx=3*frand();
        bool success=quadraticsolver(a, b, c, mn, mx, false);
        if (! success)
          {
            printf("Test %d (%f,%f,%f,%f,%f) failed\n", t, a,b,c,mn,mx);
            good=false;
          }
      }
    return good;
  }

bool Contour1DTest()
  {
    bool good=PerformContour1DTest(SolveQuadratic_MaxD, "MaxD");
    good=PerformContour1DTest(SolveQuadratic, "Cubic") && good;
    // Linear solver can't find double roots, so don't include it
    // in "good"
    PerformContour1DTest(SolveQuadratic_Linear, "Linear");
    return good;
  }

#if defined(PROFILE_HOOKS) && defined(PROFILE_1D)

struct ProfileData1D
  {
    ScalarType mMaxError;
    ScalarType mAvgError;
    int mContourSteps;
    int mElementsContoured;
    int mFCount;
    int mGCount;
  };

static int sIterations;
static ScalarType sNumRoots;
static ScalarType sRoots[32];

struct ErrorData1D
  {
    ScalarType mMax;
    ScalarType mAvg;
  };

void ComputeErrorSilent1D(const std::list<DomainType1d> &results, ErrorData1D *errordata)
  {
    typedef std::list<DomainType1d>::const_iterator iter;
    errordata->mMax=0;
    errordata->mAvg=0;
    int i;
    iter j;
    for (i=0; i<sNumRoots; ++i)
      {
        ScalarType mindist=1e9;
        ScalarType cur=sRoots[i];
        for (j=results.begin(); j!=results.end(); ++j)
          {
            ScalarType d=Matrix::Abs(cur-(*j)(0,0));
            if (d<mindist)
              mindist=d;
          }
        errordata->mAvg += mindist;
        if (mindist>errordata->mMax)
          errordata->mMax=mindist;
      }
    for (j=results.begin(); j!=results.end(); ++j)
      {
        ScalarType mindist=1e9;
        ScalarType cur=(*j)(0,0);
        for (i=0; i<sNumRoots; ++i)
          {
            ScalarType d=Matrix::Abs(cur-sRoots[i]);
            if (d<mindist)
              mindist=d;
          }
        errordata->mAvg += mindist;
        if (mindist>errordata->mMax)
          errordata->mMax=mindist;
      }
    errordata->mAvg /= sNumRoots+results.size();
  }

template<class Contour_T>
void ProfileContourTemplate1D(Contour_T &contour, ProfileData1D *profiledata)
  {
    for (int repeats=0; repeats<sIterations; ++repeats)
      {
        typedef Contour_T ContourType;
        typedef ContourType::ContourList Results;
        typedef Results::iterator Iter;

        ContourTypeMaxD::MeshType::PositionType l_p, u_p;
        l_p(0,0)=0;
        u_p(0,0)=1;
        contour.mMesh.Initialize(l_p, u_p);

        contour.mEpsilon = epsilon;

        Function1DProfile *func=static_cast<Function1DProfile *>
                                (contour.mInterpolant.mFunction);

        func->ResetCounts();

        contour.Init();

        profiledata->mContourSteps=0;
        Results results;
        {
          Profile_Scope(contour_once);
          {
            Profile_Scope(mesh_refinement);

            bool workleft;
            do
              {
                workleft = contour.Step();
                ++profiledata->mContourSteps;
              } while (workleft);
          }

          {
            profiledata->mElementsContoured=contour.mMesh.ResultSetSize();
            {
              Profile_Scope(process_results);
              contour.Results(results);
            }
          }
        }

        profiledata->mFCount=func->mFCount;
        profiledata->mGCount=func->mGCount;

        {
          ErrorData1D errordata;
          ComputeErrorSilent1D(results, &errordata);
          profiledata->mAvgError = errordata.mAvg;
          profiledata->mMaxError = errordata.mMax;
        }
      }
  }

typedef void (*ProfileFunc)(Function1DProfile &, ProfileData1D *);

static void Profile1DMaxD(Function1DProfile &f, ProfileData1D *profiledata)
  {
    typedef ContourTypeMaxD ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;

    ProfileContourTemplate1D(contour, profiledata);
  }

static void Profile1DCubic(Function1DProfile &f, ProfileData1D *profiledata)
  {
    typedef ContourTypeReg ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;
    contour.mInterpolant.mConservativeError=2;

    ProfileContourTemplate1D(contour, profiledata);
  }

static void Profile1DLinear(Function1DProfile &f, ProfileData1D *profiledata)
  {
    typedef ContourTypeLinear ContourType;

    ContourType contour;
    contour.mInterpolant.mFunction = &f;
    contour.mInterpolant.mConservativeError=3;
    contour.mInterpolant.mThirdOrderError=800000;

    ProfileContourTemplate1D(contour, profiledata);
  }

static ProfileFunc sMethodsToProfile[] =
  {
    Profile1DMaxD, Profile1DCubic, Profile1DLinear
  };

const int cNumMethods=sizeof(sMethodsToProfile)/sizeof(sMethodsToProfile[0]);

static const char * sMethodNames[cNumMethods] =
  {
    "MaxD", "Cubic", "Linear"
  };

static void Profile1D(Function1DProfile &f, Profiler &profiler)
  {
    for (int method=0; method<cNumMethods; ++method)
      {
        printf("#%s Method ----------------------\n", sMethodNames[method]);
        printf("# Eps, #F, #G, MaxErr, AvgErr, #Contour, #Elts, TimeProfile: name=(#Exec, whole, self)\n");
        for (epsilon=0.001; epsilon>0.000000001; epsilon*=.5)
          {
            ProfileData1D profiledata;
            profiler.Start();
            sMethodsToProfile[method](f, & profiledata);
            profiler.Done();
            // print report
            printf("%g, %d, %d, %g, %g, %d", 
                    epsilon, 
                    profiledata.mFCount, 
                    profiledata.mGCount,
                    profiledata.mMaxError, 
                    profiledata.mAvgError,
                    profiledata.mContourSteps,
                    profiledata.mElementsContoured);
            double maxwholetime=0;
            for (Profiler::ReportIter report_item=profiler.mReport.begin();
                  report_item!=profiler.mReport.end(); ++report_item)
              {
                printf(", %s=(%d, %f, %f)",
                  (*report_item).mName, (*report_item).mNumExecutions, 
                  (*report_item).mWholeTime, (*report_item).mSelfTime);
                if ((*report_item).mWholeTime>maxwholetime)
                  maxwholetime=(*report_item).mWholeTime;
              }
            printf("\n");
            fflush(stdout);
            if (maxwholetime>=1000)
              break;
          }
        printf("\n");
      }
  }

void ProfileContour1D()
  {
    std::auto_ptr<Profiler> profiler(Profiler::Create());
    {
      SinXXFunction f;
      printf("#******************************** sin(100x^2)/x \n");

      sIterations=100;
      sNumRoots=32;
      for (int i=0; i<32; ++i)
        sRoots[i]=sqrt(i*3.1415926535897932384626)/10;

      Profile1D(f, *profiler);
    }
  }

#endif