#include "root1d.h"

#include <vector>
#include <algorithm>
#include <boost/scoped_array.hpp>
#include "smallmat.h"
#include "matrix.h"

namespace Root1D
  {
    
    SignResult SignOfProduct(ScalarType a, ScalarType b)
      {
        ScalarType prod = a*b;
        if (prod<0)
          return Negative;
        else if (prod>0)
          return Positive;
        else if (a==0)
          return Zero;
        else if (b==0)
          return Zero;
        // deal with underflow of product
        else if (a>0)
          {
            if (b>0)
              return Positive;
            else if (b<0)
              return Negative;
          }
        else if (a<0)
          {
            if (b>0)
              return Negative;
            else if (b<0)
              return Positive;
          }
        assert(0);
        return Zero;
      }

    
    Hermite ::Hermite
        (ScalarType x0, ScalarType xd, 
         ScalarType y0, ScalarType y1, 
         ScalarType d0, ScalarType d1)
        : mX0(x0), mXd(xd)
      {
        mY[0]=y0;
        mY[1]=y1;
        mD[0]=d0;
        mD[1]=d1;
      }


     // t in 0..1
    
    void BezierUnit ::EvaluateAt(ScalarType t, ScalarType &y, ScalarType &d) const
      {
        ScalarType yt[3];
        yt[0] = mY[0] + (mY[1]-mY[0])*t;  // at t/3
        yt[1] = mY[1] + (mY[2]-mY[1])*t;  // at 1/3 + t/3
        yt[2] = mY[2] + (mY[3]-mY[2])*t;  // at 2/3 + t/3
        yt[0] = yt[0] + (yt[1]-yt[0])*t;  // at 2*t/3
        yt[1] = yt[1] + (yt[2]-yt[1])*t;  // at 1/3 + 2*t/3
        yt[2] = yt[1]-yt[0];              // length = 1/3
        y = yt[0] + yt[2]*t;
        d = yt[2] * 3;
      }

    
    ScalarType BezierUnit ::EvaluateAt(ScalarType t) const
      {
        ScalarType yt[3];
        yt[0] = mY[0] + (mY[1]-mY[0])*t;  // at t/3
        yt[1] = mY[1] + (mY[2]-mY[1])*t;  // at 1/3 + t/3
        yt[2] = mY[2] + (mY[3]-mY[2])*t;  // at 2/3 + t/3
        yt[0] = yt[0] + (yt[1]-yt[0])*t;  // at 2*t/3
        yt[1] = yt[1] + (yt[2]-yt[1])*t;  // at 1/3 + 2*t/3
        return yt[0] + (yt[1]-yt[0])*t;
      }

     // t in 0..1 corresponding to mX0..mX0+mXd
    
    void BezierInterval ::Subdivide(ScalarType t, BezierInterval  &b)
      { // De Casteljau subdivision
        ScalarType yt[6];
        yt[0] = mY[0] + (mY[1]-mY[0])*t;     // at t/3
        yt[1] = mY[1] + (mY[2]-mY[1])*t;     // at 1/3 + t/3
        yt[2] = mY[2] + (mY[3]-mY[2])*t;     // at 2/3 + t/3
        yt[3] = yt[0] + (yt[1]-yt[0])*t;  // at 2*t/3
        yt[4] = yt[1] + (yt[2]-yt[1])*t;  // at 1/3 + 2*t/3
        yt[5] = yt[3] + (yt[4]-yt[3])*t;  // at t
        ScalarType nxd=t*mXd;
        b.mX0 = mX0+nxd;
        b.mXd = mXd-nxd;
        mXd=nxd;
        b.mY[0] = yt[5];
        b.mY[1] = yt[4];
        b.mY[2] = yt[2];
        b.mY[3] = mY[3];
        mY[1] = yt[0];
        mY[2] = yt[3];
        mY[3] = yt[5];
      }

    
    void BezierInterval ::EvaluateAt(ScalarType t, ScalarType &y, ScalarType &d) const
      {
        ScalarType yt[3];
        yt[0] = mY[0] + (mY[1]-mY[0])*t;  // at t/3
        yt[1] = mY[1] + (mY[2]-mY[1])*t;  // at 1/3 + t/3
        yt[2] = mY[2] + (mY[3]-mY[2])*t;  // at 2/3 + t/3
        yt[0] = yt[0] + (yt[1]-yt[0])*t;  // at 2*t/3
        yt[1] = yt[1] + (yt[2]-yt[1])*t;  // at 1/3 + 2*t/3
        yt[2] = yt[1]-yt[0];              // length = mXd/3
        y = yt[0] + yt[2]*t;
        d = yt[2] * 3. / mXd;
      }

    // 0 <= t0 <= t1 <= 1
    
    void BezierInterval ::Subsegment(ScalarType t0, ScalarType t1, Hermite  &h) const
      { // determine the value and derivative at both points,
        // then find corresponding Bezier
        EvaluateAt(t0, h.mY[0], h.mD[0]);
        EvaluateAt(t1, h.mY[1], h.mD[1]);
        h.mX0 = mX0+t0*mXd;
        h.mXd = (t1-t0)*mXd;
      }

    
    void BezierNthDegreeUnit ::copy(const self & cpy)
      {
        mN=cpy.mN;
        mY=new ScalarType[mN];
        for (int i=0; i<=mN; ++i)
          mY[i]=cpy.mY[i];
      }

    
    void BezierNthDegreeUnit ::destroy()
      {
        if (mY)
          {
            delete[] mY;
            mY=0;
          }
      }

    // To evaluate (m j) (n k-j) / (m+n k) 
    //  = m! n! k! m+n-k!/ (j! m-j! k-j! n-k+j! m+n!)
    
    ScalarType BezierNthDegreeUnit 
    ::BigChoose(int m, int n, int j, int k)
      {
        int *factors=new int[m+n];
        int i;
          // / m+n!
        for (i=1; i<m+n; ++i)
          factors[i]=-1;
          // * m!
        for (i=1; i<m; ++i)
          ++factors[i];
          // * n!
        for (i=1; i<n; ++i)
          ++factors[i];
          // * k!
        for (i=1; i<k; ++i)
          ++factors[i];
          // * m+n-k
        for (i=1; i<m+n-k; ++i)
          ++factors[i];
          // / j!
        for (i=1; i<j; ++i)
          --factors[i];
          // / m-j!
        for (i=1; i<m-j; ++i)
          --factors[i];
          // / k-j!
        for (i=1; i<k-j; ++i)
          --factors[i];
          // / n-k+j!
        for (i=1; i<n-k+j; ++i)
          --factors[i];

        ScalarType ret=1;
        for (i=1; i<m+n; ++i)
          if (factors[i]>0)
            {
              do
                {
                  ret *= i+1;
                } while (--factors[i]);
            }
          else if (factors[i]<0)
            {
              do
                {
                  ret /= i+1;
                } while (++factors[i]);
            }

        delete[] factors;
        return ret;
      }

    
    void BezierNthDegreeUnit 
    ::Multiply(ScalarType *outY, int outN, 
               ScalarType *lY,   int lN, 
               ScalarType *rY,   int rN)
      {
        assert(outN==lN+rN);
        for (int k=0; k<=outN; ++k)
          {
            outY[k]=0;
            for (int j=Matrix::Max(0, k-rN); j<=Matrix::Min(lN, k); ++j)
              outY[k] += BigChoose(lN, rN, j, k)*lY[j]*rY[k-j];
          }
      }

    
    void BezierNthDegreeUnit ::operator+=(const self & rhs)
      {
        assert(mN==rhs.mN);
        for (int i=0; i<=mN; ++i)
          mY[i] += rhs.mY[i];
      }

    
    void BezierNthDegreeUnit ::operator-=(const self & rhs)
      {
        assert(mN==rhs.mN);
        for (int i=0; i<=mN; ++i)
          mY[i] -= rhs.mY[i];
      }

    
    void BezierNthDegreeUnit ::operator*=(ScalarType rhs)
      {
        for (int i=0; i<=mN; ++i)
          mY[i] *= rhs;
      }

    
    void BezierNthDegreeUnit ::operator*=(const self & rhs)
      {
        int newN=mN+rhs.mN;
        ScalarType *newY=new ScalarType[newN+1];
        Multiply(newY, newN, mY, mN, rhs.mY, rhs.mN);
        destroy();
        mY=newY;
        mN=newN;
      }

    
    BezierNthDegreeUnit  
    BezierNthDegreeUnit 
    ::operator*(const self & r)
      {
        self ret(mN+r.mN);
        Multiply(ret.mY, ret.mN, mY, mN, r.mY, r.mN);
        return ret;
      }

    
    void BezierNthDegreeUnit 
    ::Rescale()
      {
        ScalarType maxval=0;
        int i;
        for (i=0; i<=mN; ++i)
          {
            ScalarType absval=Matrix::Abs(mY[i]);
            if (absval>maxval)
              maxval=absval;
          }
        if (maxval<1 && maxval>0)
          {
            for (i=0; i<=mN; ++i)
              mY[i]/=maxval;
          }
      }
    
    // Simplest bisection version of cubic root finding
    void Solve_Bisection(std::list<DomainType1d> &results, 
                        ScalarType x0, ScalarType y0, ScalarType d0,
                        ScalarType xd, ScalarType y1, ScalarType d1,
                        ScalarType epsilon)
      {
        ScalarType y13 = y0 + d0*xd/3.;
        ScalarType y23 = y1 - d1*xd/3.;
        bool pos = y0>0;
        // check if all positive or all negative
        if (pos==(y13>0) && pos==(y23>0) && pos==(y1>0))
          // convex hull of control polygon does not intersect x-axis
          // so no roots
          return;

        if (Matrix::Abs(xd) < epsilon)
          { // interval small enough, assume we found a root
            AddResult(results, x0+xd/2);
            return;
          }

        ScalarType y12;
        ScalarType d12;
        // subdivide in half
        y12=(y0+y1)/2+(d0-d1)*xd/8;
        d12=((y1-y0)*3/xd) - (d0+d1);
        xd /= 2;

        if (y12==0)
          AddResult(results, x0 + xd);
        Solve_Bisection(results, x0, y0, d0, xd, y12, d12, epsilon);
        Solve_Bisection(results, x0+xd, y12, d12, xd, y1, d1, epsilon);
      }

    // Simplest bisection version of cubic root finding, Hermite version
    
    void Solve_Bisection(std::list<DomainType1d> &results, 
                         Hermite  h, ScalarType epsilon)
      {
        while (1) // tail recursion elimination
          {
            BezierUnit b;
            b.SetFromHermite(h);
            // check if all positive or all negative
            bool pos = b.mY[0]>0;
            if (pos==(b.mY[1]>0) && pos==(b.mY[2]>0) && pos==(b.mY[3]>0))
              // convex hull of control polygon does not intersect x-axis
              // so no roots
              return;
      
            if (Matrix::Abs(h.mXd) < epsilon)
              { // interval small enough, assume we found a root
                AddResult(results, h.mX0+h.mXd/2);
                return;
              }

            ScalarType y12;
            ScalarType d12;
            // subdivide in half
            y12=(h.mY[0]+h.mY[1])/2+(h.mD[0]-h.mD[1])*h.mXd/8;
            d12=((h.mY[1]-h.mY[0])*3/h.mXd) - (h.mD[0]+h.mD[1]);
            h.mXd /= 2;

            if (y12==0)
              AddResult(results, h.mX0 + h.mXd);

            Solve_Bisection(results, Hermite(h.mX0+h.mXd, h.mXd, y12, h.mY[1], d12, h.mD[1]), epsilon);
            // update h for next iteration of loop
            h.mY[1] = y12;
            h.mD[2] = d12;
          }
      }

    // Simplest bisection version of cubic root finding, Bezier version
    
    void Solve_Bisection(std::list<DomainType1d> &results, 
                         BezierInterval  b, ScalarType epsilon)
      {
        while (1) // tail recursion elimination
          {
            // check if all positive or all negative
            if (b.mY[0]>0 || (b.mY[0]==0 && (b.mY[1]>0 || (b.mY[1]==0 && 
                                            (b.mY[2]>0 || (b.mY[2]==0 && b.mY[3]>=0))))))
              {
                if (b.mY[1]>=0 && b.mY[2]>=0 && b.mY[3]>=0)
                  // convex hull of control polygon does not intersect x-axis
                  // so no roots
                  return;
              }
            else
              {
                if (b.mY[1]<=0 && b.mY[2]<=0 && b.mY[3]<=0)
                  // convex hull of control polygon does not intersect x-axis
                  // so no roots
                  return;
              }
      
            if (Matrix::Abs(b.mXd) < epsilon)
              { // interval small enough, assume we found a root
                AddResult(results, b.mX0+b.mXd/2);
                return;
              }

            BezierInterval  b2;
            // subdivide in half
            b.Subdivide(.5, b2);
            // b is first half, b2 is second half

            if (b2.mY[0]==0)
              AddResult(results, b2.mX0);

            Solve_Bisection(results, b2, epsilon);
            // use new b for next iteration of loop
          }
      }


    // Helper routines for HullIntersection functions ------------------------
    
    void EnlargeExtents(std::vector<VerticalSegment  > &hull, 
                        ScalarType epsilonR)
      {
        for (int i=0; i<hull.size(); ++i)
          {
            hull[i].mY[0]-=epsilonR;
            hull[i].mY[1]+=epsilonR;
          }
      }

    // HullInterseciton(...) -----------------------------------------------
    // returns false if no intersection
    
    bool HullIntersection(const std::vector<VerticalSegment  > &hull,
                          ScalarType & minX, ScalarType & maxX)
      {
        int sz = hull.size();
        assert(sz>1);
        std::vector<ScalarType> t;
        t.reserve(sz*(sz-1)+2);
        if (hull.front().mY[0]<=0 && hull.front().mY[1]>=0)
          t.push_back(0);

        if (hull.back().mY[0]<=0 && hull.back().mY[1]>=0)
          t.push_back(1);
        
        ScalarType xd = ScalarType(1)/(sz-1);
        for (int j=0; j<2; ++j)
          for (int i=0; i<sz-1; ++i)
            for (int k=i+1; k<sz; ++k)
              {
                ScalarType intersection;
                if (Root1D::LineSegmentIntersection(hull[i].mY[j], 
                                      hull[k].mY[j], xd*(k-i), intersection))
                  {
                    intersection += i*xd;
                    t.push_back(intersection);
                  }
              }
        if (t.empty())
          return false;
        Root1D::FindMinMax(t, minX, maxX);
        return true;
      };

    // Helper routines for HullIntersection_Fast -----------------------------

    typedef enum { Below, Cross, Above } CaseType;

    
    CaseType ClassifyVerticalSegment(const VerticalSegment  &segment)
      {
        if (segment.mY[1] < 0)
          return Below;
        else if (segment.mY[0] > 0)
          return Above;
        else
          return Cross;
      }

    
    void FindMinIntersection(const std::vector<VerticalSegment  >
                                &hull, int which, ScalarType & minX)
      {
        int sz = hull.size();
        assert(sz>1);
        minX = sz+1;
        int firstmax = sz-1;
        for (int first=0; first<firstmax; ++first)
          for (int second=first+1; second<sz; ++second)
            {
              ScalarType intersection;
              if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                  hull[second].mY[which], ScalarType(second-first), intersection))
                {
                  intersection += first;
                  if (intersection<minX)
                    {
                      minX=intersection;
                      firstmax = Matrix::Min(firstmax, static_cast<int>(intersection)+1);
                    }
                }
            }
        assert(minX<=sz);
        minX /= (sz-1);
      }

    
    void FindMaxIntersection(const std::vector<VerticalSegment  >
                                &hull, int which, ScalarType & maxX)
      {
        int sz = hull.size();
        assert(sz>1);
        maxX = -1;
        int secondmin = 0;
        for (int second=sz-1; second>secondmin; --second)
          for (int first=second-1; first>=0; --first)
            {
              ScalarType intersection;
              if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                  hull[second].mY[which], ScalarType(second-first), intersection))
                {
                  intersection += first;
                  if (intersection>maxX)
                    {
                      maxX=intersection;
                      secondmin = static_cast<int>(intersection);
                    }
                }
            }
        assert(maxX>=0);
        maxX /= (sz-1);
      }

    // returns false if no intersection
    
    bool FindBothIntersections(const std::vector<VerticalSegment  >
                                &hull, int which, ScalarType & minX, ScalarType & maxX)
      { // if which == 1, hull.front() and hull.back() are below
        // if which == 0, hull.front() and hull.back() are above
        // in either case, if there is an intersection it will be from
        // hull[..].mY[which]

        int sz = hull.size();
        if (sz<=2)
          return false;
        int i;
        for (i=1; i<sz-1; ++i)
          if (hull[i].mY[which]==0 || (hull[i].mY[which]>0) == which)
            break;
        if (i==sz-1)
          // did not cross x-axis
          return false;
        int j;
        for (j=sz-2; j>i; --j)
          if (hull[j].mY[which]==0 || (hull[j].mY[which]>0) == which)
            break;
        ++j;

        // 0 .. i-1        j..sz-1
        // ------------------------- x-axis (or flipped)
        //          i..j-1

        if (j-i<=2)
          { // at most two sign changes
            // Compute minX
            {
              minX = sz+1;
              for (int first=0; first<i; ++first)
                for (int second=i; second<j; ++second)
                  {
                    ScalarType intersection;
                    if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                        hull[second].mY[which], ScalarType(second-first), intersection))
                      {
                        intersection += first;
                        if (intersection < minX)
                          minX=intersection;
                      }
                  }
              assert(minX<=sz);
              minX /= (sz-1);
            }

            // Compute maxX
            {
              maxX = -1;
              for (int first=i; first<j; ++first)
                for (int second=j; second<sz; ++second)
                  {
                    ScalarType intersection;
                    if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                        hull[second].mY[which], ScalarType(second-first), intersection))
                      {
                        intersection += first;
                        if (intersection > maxX)
                          maxX=intersection;
                      }
                  }
              assert(maxX>=0);
              maxX /= (sz-1);
            }
          }
        else
          {
            {
              minX = sz+1;
              for (int first=0; first<i; ++first)
                for (int second=i; second<j; ++second)
                  if (SignOfProduct(hull[first].mY[which], 
                                    hull[second].mY[which])!=Positive)
                    {
                      ScalarType intersection;
                      if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                          hull[second].mY[which], ScalarType(second-first), intersection))
                        {
                          intersection += first;
                          if (intersection < minX)
                            minX=intersection;
                        }
                    }
              assert(minX<=sz);
              minX /= (sz-1);
            }

            // Compute maxX
            {
              maxX = -1;
              for (int first=i; first<j; ++first)
                for (int second=j; second<sz; ++second)
                  if (SignOfProduct(hull[first].mY[which], 
                                    hull[second].mY[which])!=Positive)
                    {
                      ScalarType intersection;
                      if (Root1D::LineSegmentIntersection(hull[first].mY[which], 
                          hull[second].mY[which], ScalarType(second-first), intersection))
                        {
                          intersection += first;
                          if (intersection > maxX)
                            maxX=intersection;
                        }
                    }
              assert(maxX>=0);
              maxX /= (sz-1);
            }
          }

        return true;
      }

    // HullInterseciton_Fast(...) --------------------------------------------
    // returns false if no intersection
    
    bool HullIntersection_Fast(const std::vector<VerticalSegment 
                               > &hull, 
                              ScalarType & minX, ScalarType & maxX)
      {
        // Consider the first and last vertical segment
        // either they are above, below, or cross the x-axis
        unsigned cases[2] = { ClassifyVerticalSegment(hull.front()), 
                              ClassifyVerticalSegment(hull.back()) };

        // cases[0]+3*cases[1]
        switch (cases[0]+cases[1]+(cases[1]<<1))
          {
            //    both cross => hull = [0,1]
            case Cross+3*Cross:
              minX=0;
              maxX=1;
              return true;

            //    both above or both below => maybe no hull, 
            //          see if any min/max has the other sign, 
            //          only need to consider one of top/bottom line
            // (here we assume hull has size at most 4 when finding roots)
            case Below+3*Below:
              return FindBothIntersections(hull, 1, minX, maxX);

            case Above+3*Above:
              return FindBothIntersections(hull, 0, minX, maxX);

            //    one above, one below => bottom line will give either max or 
            //          min, top line will give other
            case Below+3*Above:
              FindMinIntersection(hull, 1, minX);
              FindMaxIntersection(hull, 0, maxX);
              return true;

            case Above+3*Below:
              FindMinIntersection(hull, 0, minX);
              FindMaxIntersection(hull, 1, maxX);
              return true;

            //    one crosses, one above/below => know either max or min
            //          also know if bottom or top line will give the other
            case Cross+3*Below:
              minX=0;
              FindMaxIntersection(hull, 1, maxX);
              return true;

            case Cross+3*Above:
              minX=0;
              FindMaxIntersection(hull, 0, maxX);
              return true;

            case Below+3*Cross:
              FindMinIntersection(hull, 1, minX);
              maxX=1;
              return true;

            case Above+3*Cross:
              FindMinIntersection(hull, 0, minX);
              maxX=1;
              return true;
          }
        assert(0);  // should not get here
        return false;
      };

    // --------------- Quadratic convergence method -----------------
    // Version of cubic root finding with quadratic convergence
    // Sherbrooke and Patrikalakis 1993 (CAGD 10 pp. 379-405)
    // Uses intersection of convex hull of control polygon with
    // x-axis to determine new interval
    
    void Solve_Quadratic(std::list<ScalarType> &results, 
                         BezierInterval  b, 
                         ScalarType epsilon, ScalarType epsilonR)
      {
        while (1) // tail recursion elimination
          {
            // check if all positive or all negative
            if (b.mY[0]>epsilonR && b.mY[1]>epsilonR && 
                b.mY[2]>epsilonR && b.mY[3]>epsilonR)
              // convex hull of control polygon does not intersect x-axis
              // so no roots
              return;
            if (b.mY[0]<-epsilonR && b.mY[1]<-epsilonR && 
                b.mY[2]<-epsilonR && b.mY[3]<-epsilonR)
              // convex hull of control polygon does not intersect x-axis
              // so no roots
              return;
      
            if (Matrix::Abs(b.mXd) <= epsilon)
              { // interval small enough, assume we found a root
                results.push_back(b.mX0+b.mXd/2);
                return;
              }

            // compute convex hull of control polygon
            // and intersect the results with the x-axis
            ScalarType mint, maxt;
            {
              std::vector<VerticalSegment  > extents;
              extents.reserve(4);
              for (int i=0; i<4; ++i)
                {
                  extents.push_back();
                  extents.back().mY[0]=b.mY[i]-epsilonR;
                  extents.back().mY[1]=b.mY[i]+epsilonR;
                }
              if (!HullIntersection_Fast(extents, mint, maxt))
                return;
            }

            // avoid missing the root when (almost) linear
            if ((maxt-mint)*b.mXd < epsilon)
              {
                ScalarType addto = epsilon/b.mXd - (maxt-mint);
                addto /= 3;
                mint -= addto;
                maxt += addto;

                Hermite  h;
                b.Subsegment(mint, maxt, h);
                b.SetFromHermite(h);
              }
            else
              {
                Hermite  h;
                b.Subsegment(mint, maxt, h);
                b.SetFromHermite(h);
            
                if (maxt-mint > .75)
                  {
                    // need to bisect
                    BezierInterval  b2;
                    // subdivide in half
                    b.Subdivide(.5, b2);
                    // b is first half, b2 is second half
            
                    // if root at boundary of b and b2
                    // will find that root twice (probably should
                    // check explicitly for the root and move the
                    // interval endpoints inwards if it is there)
                    //Solve_Quadratic(results, b, epsilon, epsilonR);
                    Solve_Quadratic(results, b2, epsilon, epsilonR);
                    //return;
                  }
              }

            // use new b for next iteration of loop
          }
      }

    
    void Solve_Quadratic1D(std::list<ScalarType> &results, 
                            const BezierUnit  &b,
                            ScalarType epsilon, ScalarType epsilonR)
      {
        BezierInterval  b2;
        for (int i=0; i<4; ++i)
          b2.mY[i]=b.mY[i];
        b2.mXd=1;
        b2.mX0=0;

        Root1D::Solve_Quadratic(results, b2, epsilon, epsilonR);
      }

    
    void Solve_QuadraticBox(std::list<DomainType1d> &results, 
                            const BezierInterval  &b,
                            ScalarType epsilon, ScalarType epsilonR)
      {
        typedef std::list<ScalarType> RootList;
        typedef RootList::iterator    RootIter;
        RootList roots;
        Root1D::Solve_Quadratic(roots, b, epsilon, epsilonR);
        for (RootIter cur=roots.begin(); cur!=roots.end(); ++cur)
          AddResult(results, *cur);
      }

    // --------------------- SAFER quadratic method ----------------------
    
    int CountIntersections(const std::vector<VerticalSegment  >
                                &hull)
      {
        typedef std::vector<VerticalSegment  > SegList;
        typedef SegList::const_iterator SegCIter;
        int count=0;
        SegCIter segment=hull.begin();
        CaseType first=ClassifyVerticalSegment(*segment);
        CaseType second;

        for (++segment; segment!=hull.end(); ++segment)
          {
            second=ClassifyVerticalSegment(*segment);
            if (first!=second || first==Cross)
              ++count;
            first=second;
          }
        return count;
      }

    
    void ReduceNumRoots(std::list<ScalarType> &results, int maxroots,
                        BezierInterval  b)
      {
        typedef std::list<ScalarType> ResultType;
        typedef ResultType::iterator  ResultIter;
        if (results.size()>maxroots)
          {
            results.sort();
            while (results.size()>maxroots)
              { // merge the closest roots
                // maxroots >= 1, so tempresults.size() >= 2
                ResultIter closest=results.begin();
                ResultIter next=closest;
                ++next;
                ScalarType closedist=*next-*closest;
                assert(closedist>=0);
                ResultIter cur=next;
                for (++next; next!=results.end(); 
                    cur=next, ++next)
                  {
                    ScalarType dist=*next-*cur;
                    assert(dist>=0);
                    if (dist<closedist)
                      {
                        closest=cur;
                        closedist=dist;
                      }
                  }
                next=closest;
                ++next;
                // merge *closest and *next
                // lets evaluate the function in an effort
                // to do a good job of finding the root
                ScalarType y[2];
                ScalarType d;
                b.EvaluateAt((*closest-b.mX0)/b.mXd, y[0], d);
                b.EvaluateAt((*next-b.mX0)/b.mXd,    y[1], d);
                y[0]=Matrix::Abs(y[0]);
                y[1]=Matrix::Abs(y[1]);
                ScalarType sum=y[0]+y[1];
                if (sum>0)
                  d=(*closest*y[1]+*next*y[0])/sum;
                else
                  d=(*closest+*next)*.5;
                results.erase(next);
                *closest=d;
              }
          }
      }

    // Safer version counts max number of root that can occur
    // in an interval and ensures that it won't be exceeded
    // (brute force, does not address cause; deflation method
    // below addresses the cause, but does not explicitly check
    // how many roots it generates)
    
    void Solve_Quadratic_Safer(std::list<ScalarType> &results, 
                         BezierInterval  b, 
                         ScalarType epsilon, ScalarType epsilonR)
      {
        int origmaxroots;
        {
          std::vector<VerticalSegment  > extents;
          extents.reserve(4);
          for (int i=0; i<4; ++i)
            {
              extents.push_back();
              extents.back().mY[0]=b.mY[i]-epsilonR;
              extents.back().mY[1]=b.mY[i]+epsilonR;
            }
          origmaxroots=CountIntersections(extents);
          if (origmaxroots==0)
            return;
        }

        typedef std::list<ScalarType> ResultType;

        while (1) // tail recursion elimination
          {
            // check if all positive or all negative
            if (b.mY[0]>epsilonR && b.mY[1]>epsilonR && 
                b.mY[2]>epsilonR && b.mY[3]>epsilonR)
              // convex hull of control polygon does not intersect x-axis
              // so no roots
              return;
            if (b.mY[0]<-epsilonR && b.mY[1]<-epsilonR && 
                b.mY[2]<-epsilonR && b.mY[3]<-epsilonR)
              // convex hull of control polygon does not intersect x-axis
              // so no roots
              return;
      
            if (Matrix::Abs(b.mXd) <= epsilon)
              { // interval small enough, assume we found a root
                results.push_back(b.mX0+b.mXd/2);
                return;
              }

            // compute convex hull of control polygon
            // and intersect the results with the x-axis
            ScalarType mint, maxt;
            {
              std::vector<VerticalSegment  > extents;
              extents.reserve(4);
              for (int i=0; i<4; ++i)
                {
                  extents.push_back();
                  extents.back().mY[0]=b.mY[i]-epsilonR;
                  extents.back().mY[1]=b.mY[i]+epsilonR;
                }
              if (!HullIntersection_Fast(extents, mint, maxt))
                return;
            }

            // avoid missing the root when (almost) linear
            if ((maxt-mint)*b.mXd < epsilon)
              {
                ScalarType addto = epsilon/b.mXd - (maxt-mint);
                addto /= 3;
                mint -= addto;
                maxt += addto;
                // compute new b from mint to maxt
                Hermite  h;
                b.Subsegment(mint, maxt, h);
                b.SetFromHermite(h);
              }
            else
              {
                // compute new b from mint to maxt
                Hermite  h;
                b.Subsegment(mint, maxt, h);
                b.SetFromHermite(h);
                if (maxt-mint > .75)
                  {
                    std::vector<VerticalSegment  > extents;
                    extents.reserve(4);
                    for (int i=0; i<4; ++i)
                      {
                        extents.push_back();
                        extents.back().mY[0]=b.mY[i]-epsilonR;
                        extents.back().mY[1]=b.mY[i]+epsilonR;
                      }
                    int maxroots=CountIntersections(extents);
                    maxroots=Matrix::Min(origmaxroots, maxroots);
                    if (maxroots)
                      {
                        // need to bisect
                        BezierInterval  b2;
                        // subdivide in half
                        b.Subdivide(.5, b2);
                        // b is first half, b2 is second half
                
                        ResultType tempresults;
                        // if root at boundary of b and b2
                        // will find that root twice
                        Solve_Quadratic_Safer(tempresults, b, epsilon, epsilonR);
                        Solve_Quadratic_Safer(tempresults, b2, epsilon, epsilonR);
                        ReduceNumRoots(tempresults, maxroots, b);
                        results.splice(results.end(), tempresults);
                        return;
                      }
                  }
              }

            // use new b for next iteration of loop
          }
      }

    // --------------- N deflation method -----------------------------

    // y length n+1, out length n
    
    inline void deCasteljau_N(int n, ScalarType x, const ScalarType *y,
                              ScalarType *out)
      {
        for (int i=0; i<n; ++i)
          out[i]=y[i]+x*(y[i+1]-y[i]);
      }

    // y length n+1, out length n
    
    void DivideOutRoot(int n, ScalarType root, 
                       const ScalarType *y, ScalarType *out)
      {
        assert(n>1);
        switch (n)
          {
            case 2:
              { // first subdivide at x:
                ScalarType temp[2];
                deCasteljau_N(2, root, y, temp);
                //assert(Matrix::Abs(temp[0]+(temp[1]-temp[0])*root)<0.00001);
                // output a linear function
                out[0]=2*temp[1];   // /(1-x)
                out[1]=y[2];        // /(1-x)
              }
              break;
            case 3:
              { // first subdivide
                ScalarType temp[5];
                deCasteljau_N(3, root, y, temp);
                deCasteljau_N(2, root, temp, temp+3);
                //assert(Matrix::Abs(temp[3]+(temp[4]-temp[3])*root)<0.00001);
                // output two quadratic functions
                               // C(n,i) / C(n-1,i) = n/(n-i)
                out[0]=temp[4]*3;     // /(1-x)
                out[1]=temp[2]*1.5;   // /(1-x)
                out[2]=y[3];          // /(1-x)
              }
              break;
            default:
              { // first subdivide
                boost::scoped_array<ScalarType> temp(new ScalarType[n*(n+1)/2-1]);
                deCasteljau_N(n, root, y, & temp[0]);
                {
                  int old_offset=0;
                  int offset=n;
                  for (int i=n-1; i>=2; --i)
                    {
                      deCasteljau_N(i, root, & temp[old_offset], 
                                    & temp[offset]);
                      old_offset=offset;
                      offset+=i;
                    }
                  assert(offset==(n*(n+1)/2-1));
                  //assert(Matrix::Abs(temp[offset-2]+(temp[offset-1]-
                  //                              temp[offset-2])*root)<0.00001);
                }
                // output two n-1 functions
                {
                  int firstoffs=0;
                  int len=n;
                  for (int i=1; i<n; ++i)
                    {
                      ScalarType coef=ScalarType(n)/(n-i);
                      out[n-i-1]=temp[firstoffs+len-1]*coef; // /(1-x)
                      firstoffs+=len;
                      --len;
                    }
                }
                out[n-1]=y[n];
              }
              break;
          }
      }


    // y length n+1, out length n+1
    
    void Subdivide_N(int n, ScalarType x0, ScalarType xd,
                     const ScalarType *y, ScalarType *out)
      {
        if (x0>=1)
          {
            for (int i=0; i<=n; ++i)
              out[i]=y[n];
            return;
          }
        boost::scoped_array<ScalarType> temp(new ScalarType[n*(n+1)/2]);
        deCasteljau_N(n, x0, y, &temp[0]);
        {
          int old_offset=0;
          int offset=n;
          for (int i=n-1; i>=1; --i)
            {
              deCasteljau_N(i, x0, & temp[old_offset], & temp[offset]);
              old_offset=offset;
              offset+=i;
            }
          assert(offset==(n*(n+1)/2));
        }
        boost::scoped_array<ScalarType> temp2(new ScalarType[n+1]);
        {
          int offs=0;
          int len=n;
          for (int i=1; i<=n; ++i)
            {
              temp2[n-i]=temp[offs+len-1];
              offs+=len;
              --len;
            }
        }
        temp2[n]=y[n];
        boost::scoped_array<ScalarType> temp3(new ScalarType[n*(n+1)/2]);
        ScalarType t=xd/(1-x0);
        deCasteljau_N(n, t, & temp2[0], & temp3[0]);
        {
          int old_offset=0;
          int offset=n;
          for (int i=n-1; i>=1; --i)
            {
              deCasteljau_N(i, t, & temp3[old_offset], & temp3[offset]);
              old_offset=offset;
              offset+=i;
            }
          assert(offset==(n*(n+1)/2));
        }
        out[0]=temp2[0];
        {
          int offs=0;
          int len=n;
          for (int i=1; i<=n; ++i)
            {
              out[i]=temp3[offs];
              offs+=len;
              --len;
            }
        }
      }

    // less recursive formulation
    
    void Solve_N(std::list<ScalarType> &results, int n, const ScalarType *y,
                         ScalarType x[2], 
                         ScalarType epsilon, ScalarType epsilonR)
      {
        typedef std::list<ScalarType> ResultType;
        typedef ResultType::iterator  ResultIter;

        // subdivide interval
        boost::scoped_array<ScalarType> sub(new ScalarType[n+1]);
        ScalarType curx = x[0];
        ScalarType xdelta=x[1]-x[0];
        std::vector<VerticalSegment  > extents(n+1);
        extents.reserve(n+1);
        while (true)  // tail recursion elimination
          {
            Subdivide_N(n, curx, xdelta, y, & sub[0]);

            // find hull of sub-spline
            ScalarType newt[2];
            {
              for (int i=0; i<=n; ++i)
                {
                  extents[i].mY[0]=sub[i]-epsilonR;
                  extents[i].mY[1]=sub[i]+epsilonR;
                }
              if (!HullIntersection_Fast(extents, newt[0], newt[1]))
                return;
            }

            if (xdelta < epsilon)
              {
                ScalarType tdelta=newt[1]-newt[0];
                curx+=xdelta*newt[0];
                xdelta*=tdelta;
                results.push_back(curx+xdelta*.5);
                return;
              }

            ScalarType tdelta=newt[1]-newt[0];
            curx+=xdelta*newt[0];
            xdelta*=tdelta;

            ScalarType enlargeby=epsilon-xdelta;
            if (enlargeby > 0)
              { // enlarge so don't miss roots
                enlargeby /= 3; // /=2 would give an interval length=epsilon
                curx-=enlargeby;
                xdelta+=2*enlargeby;
                // recurse (one last time) to see if this is a root
              }
            else if (tdelta > .75)
              { // subdivide in half
                ScalarType half[2];
                half[0]=curx;
                half[1]=curx+xdelta*.5;
                ResultType tempresults;
                Solve_N(tempresults, n, y, half, epsilon, epsilonR);
                if (tempresults.empty())
                  { // no root in the first half, look in the second
                    half[0]=half[1];
                    half[1]=curx+xdelta;
                    Solve_N(results, n, y, half, epsilon, epsilonR);
                  }
                else
                  { // found a root in the first half,
                    // factor it out so we don't find it in the second
                    ResultIter cur=tempresults.begin();
                    ScalarType root=*cur;
                    for (++cur; cur!=tempresults.end(); ++cur)
                      {
                        if (*cur > root)
                          root=*cur;
                      }
                    results.splice(results.end(), tempresults);
                    assert(tempresults.empty());

                    half[0]=Matrix::Max(half[1], root+epsilon);
                    half[1]=curx+xdelta;
                    if (half[0]<=half[1] && n>1)
                      {
                        ScalarType convertedx[2];
                        assert(root<1);
                        convertedx[0] = (half[0]-root)/(1-root);
                        convertedx[1] = (half[1]-root)/(1-root);
                        boost::scoped_array<ScalarType> right(new ScalarType[n]);
                        DivideOutRoot(n, root, y, & right[0]);
                        Solve_N(tempresults, n-1, & right[0],
                                 convertedx, epsilon/(1-root), epsilonR*(1-root));
                        for (cur=tempresults.begin(); cur!=tempresults.end(); 
                            ++cur)
                          results.push_back(root+(*cur*(1-root)));
                      }
                  }
                return;
              }
            // else recurse
          }
      }

    
    void Solve_N_P(std::list<ScalarType> &results, 
                           const BezierInterval  &b, 
                           ScalarType epsilon, ScalarType epsilonR)
      {
        ScalarType x[2] = { 0, 1 };
        typedef std::list<ScalarType> ResultsType;
        typedef ResultsType::iterator ResultsIter;
        ResultsType tempresults;
        if (b.mXd>epsilon)
          {
            Solve_N(tempresults, 3, b.mY, x, epsilon/b.mXd, epsilonR);
            for (ResultsIter cur=tempresults.begin(); 
                  cur!=tempresults.end(); ++cur)
              results.push_back(*cur*b.mXd+b.mX0);
          }
        else
          {
            Solve_N(tempresults, 3, b.mY, x, 1, epsilonR);
            if (! tempresults.empty())
              results.push_back(b.mX0);
          }
      }

    
    void Solve_N_Box(std::list<DomainType1d> &results, 
                           const BezierInterval  &b,
                           ScalarType epsilon, ScalarType epsilonR)
      {
        ScalarType x[2] = { 0, 1 };
        typedef std::list<ScalarType> ResultsType;
        typedef ResultsType::iterator ResultsIter;
        ResultsType tempresults;
        if (b.mXd>epsilon)
          {
            Solve_N(tempresults, 3, b.mY, x, epsilon/b.mXd, epsilonR);
            for (ResultsIter cur=tempresults.begin(); 
                  cur!=tempresults.end(); ++cur)
              AddResult(results, *cur*b.mXd+b.mX0);
          }
        else
          {
            Solve_N(tempresults, 3, b.mY, x, 1, epsilonR);
            if (! tempresults.empty())
              AddResult(results, b.mX0);
          }
      }
  }
