#include "root1d.h"
#include <stdio.h>

#ifdef TEMPLATED_CODE
typedef double scalarT;
typedef Matrix::SmallMatrix<scalarT, 1, 1> domainT;
#else
typedef ScalarType scalarT;
typedef DomainType1d domainT;
#endif

static scalarT epsilon = 0.000001;
static scalarT epsilonR = epsilon*epsilon*.5;

static scalarT frand()
  {
    return ((scalarT)rand())/(RAND_MAX+1);
  }

static bool NotNear(scalarT a, scalarT b)
  {
    scalarT diff=a-b;
    if (diff<0)
      diff= -diff;
    return diff > epsilon;
  }

static bool NotNear(scalarT a[2])
  {
    scalarT diff=a[0]-a[1];
    if (diff<0)
      diff= -diff;
    return diff > epsilon;
  }

static bool TestHullIntersection()
  {
    bool ret_good=true;
    std::vector<Root1D::VerticalSegment> v(3);
    for (int num=3; num<=10; ++num)
      {
        int iterations=32767;
        printf("Testing hull intersection with %d segments (%d iterations)\n",
                num, iterations);
        for (int test=0; test<iterations; ++test)
          {
            for (int i=0; i<num; ++i)
              {
                v[i].mY[0]=frand()-.5;
                v[i].mY[1]=frand()-.5;
                if (v[i].mY[1] < v[i].mY[0])
                  std::swap(v[i].mY[0], v[i].mY[1]);
              }
            scalarT minX[2], maxX[2];
            bool wasintersection[2];
            //for (int j=0; j<num; ++j)
            //  {
            //    printf("%f -- %f    ", v[j].mY[0], v[j].mY[1]);
            //  }
            //printf("\n");
            wasintersection[0]=Root1D::HullIntersection     (v, minX[0], maxX[0]);
            wasintersection[1]=Root1D::HullIntersection_Fast(v, minX[1], maxX[1]);
            if (wasintersection[0]!=wasintersection[1] || 
                (wasintersection[0] && (NotNear(minX) || NotNear(maxX))))
              {
                printf("HullIntersection Test %d.%d failed\n", num, test);
                ret_good=false;
              }
            //else
            //  {
            //    if (wasintersection[0])
            //      printf("true: %f %f\n", minX[0], maxX[0]);
            //    else
            //      printf("false\n");
            //  }

          }
        v.push_back();
      }
    if (ret_good)
      printf("All hull intersection tests passed\n");
    return ret_good;
  }

static void PrintScalars(const std::list<scalarT> &results)
  {
    typedef std::list<scalarT> ScalarList;
    typedef ScalarList::const_iterator  ScalarCIter;
    if (! results.empty())
      {
        ScalarCIter cur=results.begin();
        printf("%f", *cur);
        for (++cur; cur!=results.end(); ++cur)
          printf(", %f", *cur);
      }
    else
      printf("*none*");
  }

static void PrintDomain(const std::list<domainT> &results)
  {
    typedef std::list<domainT> DomainList;
    typedef DomainList::const_iterator DomainCIter;
    if (! results.empty())
      {
        DomainCIter cur=results.begin();
        printf("%f", (*cur)(0,0));
        for (++cur; cur!=results.end(); ++cur)
          printf(", %f", (*cur)(0,0));
      }
    else
      printf("*none*");
  }

namespace Matrix
  {
    bool operator<(domainT l, domainT r)
      {
        return l(0,0) < r(0,0);
      }
  }

static bool TestRootFinding()
  {
    int rootcounts[4] = {0,0,0,0};
    int i;
    bool ret=true;
    int iterations=100000;
    printf("Testing 1d root finding (%d iterations)\n", iterations);
    for (i=0; i<iterations; ++i)
      {
        Root1D::BezierInterval b;
        b.mXd=1;
        b.mX0=0;
        for (int j=0; j<4; ++j)
          b.mY[j]=frand()*2-1;
        std::list<domainT> domainlist;
        std::list<scalarT> scalarlist;
        std::list<scalarT> scalarlist2;
        if (b.mY[3]==0)
          Root1D::AddResult(domainlist, 1);
        Root1D::Solve_Bisection(domainlist, b, epsilon);
        if (b.mY[0]==0)
          Root1D::AddResult(domainlist, 0);
        domainlist.sort();

        Root1D::Solve_Quadratic(scalarlist, b, epsilon, epsilonR);
        scalarlist.reverse();

        Root1D::Solve_N_P(scalarlist2, b, epsilon, epsilonR);

        bool printit=false;
        if (domainlist.size()!=scalarlist.size() || 
            domainlist.size()!=scalarlist2.size())
          printit=true;
        else
          {
            assert(domainlist.size()<=3);
            ++rootcounts[domainlist.size()];
            std::list<domainT>::iterator di=domainlist.begin();
            std::list<scalarT>::iterator si=scalarlist.begin(),
                                        si2=scalarlist2.begin();
            for (; di!=domainlist.end(); ++di, ++si, ++si2)
              {
                if (NotNear((*di)(0,0), *si))
                  printit=true;
                else if (NotNear((*di)(0,0), *si2))
                  printit=true;
              }
          }
        if (printit)
          {
            ret=false;
            printf("1D Root Test %d:", i);
            for (int j=0; j<4; ++j)
              {
                printf(" %f", b.mY[j]);
              }
            printf("\nBisection: ");
            PrintDomain(domainlist);
            printf("\nQuadratic: ");
            PrintScalars(scalarlist);
            printf("\nN deflate: ");
            PrintScalars(scalarlist2);
            printf("\n");
          }
      }
    for (i=0; i<4; ++i)
      printf("Tests with %d roots: %d\n", i, rootcounts[i]);
    if (ret)
      printf("All 1d root tests passed\n");
    return ret;
  }

bool TestRoot1D()
  {
    bool good=TestHullIntersection();
    good=TestRootFinding() && good;
    return good;
  }
